##############################################################################
# PyLipID: A python module for analysing protein-lipid interactions
#
# Author: Wanling Song
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
##############################################################################


r"""
util module
=============
The ``util`` module contains other assisting functions:

.. currentmodule:: pylipid.util

.. autosummary::
   :toctree: generated/

   check_dir
   write_PDB
   write_pymol_script
   sparse_corrcoef
   rmsd
   get_traj_info

"""


from .directory import check_dir
from .coordinate import write_PDB
from .pymol_script import write_pymol_script
from .corrcoef import sparse_corrcoef
from .rmsd import rmsd
from .trajectory import get_traj_info


##############################################################################
# PyLipID: A python module for analysing protein-lipid interactions
#
# Author: Wanling Song
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
##############################################################################

"""This module contains functions for clustering the bound poses. """

import numpy as np
from sklearn.neighbors import NearestNeighbors
from sklearn.cluster import DBSCAN
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score
from kneebow.rotor import Rotor


__all__ = ["cluster_DBSCAN", "cluster_KMeans"]


def cluster_DBSCAN(data, eps=None, min_samples=None, metric="euclidean"):
    """Cluster data using DBSCAN.

    The the density-based spatial cluster `sklearn.cluster.DBSCAN
    <https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html>`_
    to cluster the data. If not provided by users, the distance cutoff `eps` is determined
    by the 'Knee method' which finds the distance at which a sharp change happens.

    Parameters
    ----------
    data : ndarray, shape=(n_samples, n_dims)
    eps : None or scalar, default=None
    min_samples : None or scalar, default=None
    metric : string or callable, default=’euclidean’
        The metric to use when calculating distance between instances in a feature array. If metric
        is a string or callable, it must be one of the options allowed by `sklearn.metrics.pairwise_distances`
        for its metric parameter.
    plot_dist_curve : bool, default=True

    Returns
    -------
    labels : array_like, shape=(n_samples,)
    core_sample_indices : array_like, shape=(n_clusters,)

    """
    if len(data) <= len(data[0]):
        return np.array([0 for dummy in data]), np.arange(len(data))[np.newaxis, :]
    if eps is None:
        nearest_neighbors = NearestNeighbors(n_neighbors=3)
        nearest_neighbors.fit(data)
        distances, indices = nearest_neighbors.kneighbors(data)
        distances = np.sort(distances, axis=0)[:, 1]
        data_vstacked = np.vstack([np.arange(len(distances)), distances]).T
        rotor = Rotor()
        rotor.fit_rotate(data_vstacked)
        elbow_index = rotor.get_elbow_index()
        eps = distances[elbow_index]
    if min_samples is None:
        scores = []
        for n_sample in np.arange(2, len(data)-1, 2):
            dbscan = DBSCAN(eps=eps, min_samples=n_sample, metric=metric)
            dbscan.fit(data)
            labels = dbscan.labels_
            if np.all(labels == -1):
                break
            else:
                scores.append(silhouette_score(data, labels))
        min_samples = np.arange(2, len(data)-1, 2)[np.argmax(scores)] # the highest silhouette_score.
    dbscan = DBSCAN(eps=eps, min_samples=min_samples, metric=metric)
    dbscan.fit(data)
    core_sample_indices = [[] for label in np.unique(dbscan.labels_) if label != -1]
    for core_sample_index in dbscan.core_sample_indices_:
        core_sample_indices[dbscan.labels_[core_sample_index]].append(core_sample_index)
    return dbscan.labels_, core_sample_indices


def cluster_KMeans(data, n_clusters):
    """Cluster data using KMeans.

    Parameters
    ----------
    data : ndarray, shape=(n_samples, n_dims)
    n_clusters : scalar

    Returns
    -----------
    labels : array_like, shape=(n_samples)

    """
    if len(data) < n_clusters:
        return cluster_DBSCAN(data, eps=None, min_samples=None, metric="euclidean")
    model = KMeans(n_clusters=n_clusters).fit(data)
    labels = model.predict(data)
    return labels


##############################################################################
# PyLipID: A python module for analysing protein-lipid interactions
#
# Author: Wanling Song
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
##############################################################################

"""This module contains functions that plot interactions as a function of residue index.
"""

import os
from collections import defaultdict
import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator
import numpy as np
import pandas as pd
import logomaker

__all__ = ["plot_residue_data", "plot_residue_data_logos",
           "plot_binding_site_data", "plot_surface_area", "AxisIndex"]


def plot_residue_data(residue_index, interactions, gap=200, ylabel=None,
                      fn=None, title=None, fig_close=False):
    """Plot interactions as a function of residue index
    
    Parameters
    ----------
    residue_index : list
            Residue indices in an ascending order. If a residue index is smaller than its preceding one,
            the plotting function will consider it as the start of a new chain and will plot the following
            in a new figure. A gap of less than 50 missing residues will be marked as gray areas in the figure.
    interactions : list
            Plotting values correspond to residue_index.
    gap : int, optional, default=200
            The number of missing residues in residue_index that initiate a new figure. The gap between two adjacent
            index in residue_index that is smaller than the provided value will be considered as missing residues
            and will be marked as gray areas in the figure, whereas a gap that is larger than the provided value
            will start a new figure and plot the following data in that new figure. This can help to make figures
            more compressed. The default gap is 200.
    ylabel : str, optional, default=None
            y axis label. Default is "Interactions".
    fn : str, optional, default=None
            Figure name. By default the figure is saved as "Figure_interactions.pdf" as the current
            working directory.
    title : str, optional, default=None
            Figure title.

    """
    bar_color = "#176BA0"
    if ylabel is None:
        ylabel = "Interactions"

    if fn is None:
        fn = os.path.join(os.getcwd(), "Figure_interactions.pdf")

    # check for chain breaks
    gray_areas = defaultdict(list)  # show grey area to indicate missing residues
    chain_starts = [0]  # plot in separate figures if the gap between two adjacent residues is larger than 50
    for idx in np.arange(1, len(residue_index)):
        if residue_index[idx] - residue_index[idx - 1] < 0:
            chain_starts.append(idx)
        elif residue_index[idx] - residue_index[idx - 1] > gap:
            chain_starts.append(idx)
        elif 1 < residue_index[idx] - residue_index[idx - 1] <= gap:
            gray_areas[chain_starts[-1]].append([residue_index[idx - 1] + 1, residue_index[idx] - 1])
    chain_starts.append(len(residue_index))

    # plot
    for chain_idx in np.arange(len(chain_starts[:-1])):
        df = interactions[chain_starts[chain_idx]:chain_starts[chain_idx + 1]]
        resi_selected = residue_index[chain_starts[chain_idx]:chain_starts[chain_idx + 1]]
        if 0 < len(df) <= 20:
            fig, ax = plt.subplots(1, 1, figsize=(2.8, 1.5))
            ax.xaxis.set_major_locator(MultipleLocator(5))
            ax.xaxis.set_minor_locator(MultipleLocator(1))
        elif 20 < len(df) <= 50:
            fig, ax = plt.subplots(1, 1, figsize=(3.2, 1.5))
            ax.xaxis.set_major_locator(MultipleLocator(10))
            ax.xaxis.set_minor_locator(MultipleLocator(1))
        elif 50 < len(df) <= 300:
            fig, ax = plt.subplots(1, 1, figsize=(3.8, 1.8))
            ax.xaxis.set_major_locator(MultipleLocator(50))
            ax.xaxis.set_minor_locator(MultipleLocator(10))
        elif 300 < len(df) <= 1000:
            fig, ax = plt.subplots(1, 1, figsize=(4.5, 1.8))
            ax.xaxis.set_major_locator(MultipleLocator(100))
            ax.xaxis.set_minor_locator(MultipleLocator(10))
        elif 1000 < len(df) <= 2000:
            fig, ax = plt.subplots(1, 1, figsize=(6.0, 1.8))
            ax.xaxis.set_major_locator(MultipleLocator(200))
            ax.xaxis.set_minor_locator(MultipleLocator(50))
        elif len(df) > 2000:
            fig, ax = plt.subplots(1, 1, figsize=(7.5, 1.8))
            ax.xaxis.set_major_locator(MultipleLocator(500))
            ax.xaxis.set_minor_locator(MultipleLocator(100))
        ax.bar(resi_selected, df, 1.0, linewidth=0, color=bar_color)
        # plot missing residue area
        if chain_starts[chain_idx] in gray_areas.keys():
            for gray_area in gray_areas[chain_starts[chain_idx]]:
                ax.axvspan(gray_area[0], gray_area[1], facecolor="#c0c0c0", alpha=0.3)
        # axis setting
        ax.set_ylim(0, df.max() * 1.05)
        ax.set_xlim(resi_selected.min() - 1, resi_selected.max() + 1)
        ax.set_ylabel(ylabel, fontsize=8, weight="bold")
        ax.set_xlabel("Residue Index", fontsize=8, weight="bold")
        for label in ax.xaxis.get_ticklabels() + ax.yaxis.get_ticklabels():
            plt.setp(label, fontsize=8, weight="bold")
        if title is not None:
            ax.set_title(title, fontsize=8, weight="bold")
        plt.tight_layout()
        if len(chain_starts) == 2:
            fig.savefig(fn, dpi=300)
        else:
            name, ext = os.path.splitext(fn)
            fig.savefig("{}_{}{}".format(name, chain_idx, ext), dpi=300)
        if fig_close:
            plt.close()

    return


def plot_residue_data_logos(residue_index, logos, interactions, gap=1000, letter_map=None,
                            color_scheme="chemistry", ylabel=None, title=None, fn=None, fig_close=False):
    """Plot interactions using `logomaker.Logo
    <https://logomaker.readthedocs.io/en/latest/implementation.html#logo-class>`_.

    Parameters
    -----------
    residue_index : list
            Residue indices in an ascending order. If a residue index is smaller than its preceding one,
            the plotting function will consider it as the start of a new chain and will plot the following
            in a new figure.
    logos : list of str
            Single letter logos in the corresponding order as residue_index. The height of logos in the figure
            will be determined by values given in interactions. Three-letter name of the 20 common amino acids
            are accepted and will be converted to their corresponding single-letter names in this function by
            the default. Other mappings can be defined via letter_map.
    interactions : list
            Plotting values in the corresponding order as residue_index.
    gap : int, optional, default=1000
            The number of missing residues in residue_index that starts a new figure. A gap between two adjacent
            index in residue_index that is smaller than the provided value will be considered as missing residues
            and will be marked as gray areas in the figure, whereas a gap that is larger than the provided value
            will start a new figure and plot the following data in that new figure. This can help to make figures
            more compressed. The gap needs to be greater than 1000. The default is 1000.
    letter_map : dict, optional, default=None
            A dictionary that maps provided names to single-letter logos in the form of
            {"provided name": "single_letter logo"}.
    color_scheme : str, optional, default="chemistry"
            The color scheme used by logomaker.Logo(). See
            `Color Schemes <https://logomaker.readthedocs.io/en/latest/examples.html#color-schemes>`_ for accepted values.
            Default is "chemistry".
    ylabel : str, optional, default=None
            y axis label. Default is "Interactions".
    fn : str, optional, default=None
            Figure name. By default the figure is saved as "Figure_interactions_logo.pdf" as the current
            working directory.
    fig_close : bool, optional, default=False

    """
    # single-letter dictionary
    single_letter = {'CYS': 'C', 'ASP': 'D', 'SER': 'S', 'GLN': 'Q', 'LYS': 'K',
                     'ILE': 'I', 'PRO': 'P', 'THR': 'T', 'PHE': 'F', 'ASN': 'N',
                     'GLY': 'G', 'HIS': 'H', 'LEU': 'L', 'ARG': 'R', 'TRP': 'W',
                     'ALA': 'A', 'VAL': 'V', 'GLU': 'E', 'TYR': 'Y', 'MET': 'M'}
    if letter_map is not None:
        single_letter.update(letter_map)

    logos_checked = []
    for name in logos:
        if len(name) == 1:
            logos_checked.append(name)
        else:
            logos_checked.append(single_letter[name])
    if ylabel is None:
        ylabel = "Interactions"
    if fn is None:
        fn = os.path.join(os.getcwd(), "Figure_interactions_logo.pdf")

    length = 100
    # check for chain breaks, gray_areas and axis breaks
    axis_obj = AxisIndex(residue_index, logos_checked, interactions, length, gap)
    axis_obj.sort()
    # plot
    for page_idx in axis_obj.breaks.keys():
        n_rows = len(axis_obj.breaks[page_idx])
        fig, axes = plt.subplots(n_rows, 1, figsize=(4.5, 1.3 * n_rows), sharey=True)
        plt.subplots_adjust(hspace=0.5, left=0.2)
        ymax = []
        for ax_idx, ax in enumerate(np.atleast_1d(axes)):
            resi_selected = [item[0] for item in axis_obj.breaks[page_idx][ax_idx]]
            logos_selected = [item[1] for item in axis_obj.breaks[page_idx][ax_idx]]
            interaction_selected = [item[2] for item in axis_obj.breaks[page_idx][ax_idx]]
            ymax.append(np.max(interaction_selected))
            if np.sum(interaction_selected) > 0:
                df = pd.DataFrame({"Resid": resi_selected, "Resn": logos_selected, "Data": interaction_selected})
                matrix = df.pivot(index="Resid", columns='Resn', values="Data").fillna(0)
                logomaker.Logo(matrix, color_scheme=color_scheme, ax=ax)
            if ax_idx == (n_rows - 1):
                ax.set_xlabel("Residue Index", fontsize=8, weight="bold")
            ax.xaxis.set_major_locator(MultipleLocator(20))
            ax.xaxis.set_minor_locator(MultipleLocator(1))
            ax.set_xlim(resi_selected[0] - 0.5, resi_selected[-1] + 0.5)
            ax.set_ylabel(ylabel, fontsize=8, weight="bold", va="center")
            for label in ax.xaxis.get_ticklabels() + ax.yaxis.get_ticklabels():
                plt.setp(label, fontsize=8, weight="bold")
        np.atleast_1d(axes)[-1].set_ylim(0, np.max(ymax) * 1.05)
        # plot missing areas
        if page_idx in axis_obj.gray_areas.keys():
            for item in axis_obj.gray_areas[page_idx]:
                np.atleast_1d(axes)[item[0]].axvspan(item[1], item[2], facecolor="#c0c0c0", alpha=0.3)
        if title is not None:
            np.atleast_1d(axes)[0].set_title(title, fontsize=10, weight="bold")
        plt.tight_layout()
        if len(axis_obj.breaks.keys()) == 1:
            fig.savefig(fn, dpi=300)
        else:
            name, ext = os.path.splitext(fn)
            fig.savefig("{}_{}{}".format(name, page_idx, ext), dpi=300)
        if fig_close:
            plt.close()

    return


def plot_binding_site_data(data, fig_fn, ylabel=None, title=None, fig_close=False):
    """Plot surface area in a matplotlib violin plot.

    Parameters
    ----------
    data : padnas.DataFrame
    fig_fn : str
    ylabel : str, optional, default=None
    title : str, optional, default=None
    fig_close : bool

    """
    from itertools import cycle as _cycle

    def adjacent_values(vals, q1, q3):
        upper_adjacent_value = q3 + (q3 - q1) * 1.5
        upper_adjacent_value = np.clip(upper_adjacent_value, q3, vals[-1])
        lower_adjacent_value = q1 - (q3 - q1) * 1.5
        lower_adjacent_value = np.clip(lower_adjacent_value, vals[0], q1)
        return lower_adjacent_value, upper_adjacent_value

    if ylabel is None:
        ylabel = ""
    if title is None:
        title = ""

    color_set = _cycle(plt.get_cmap("tab10").colors)
    plt.rcParams["font.size"] = 10
    plt.rcParams["font.weight"] = "bold"

    BS_names = [col for col in data.columns]
    BS_id_set = [int(name.split()[-1]) for name in BS_names]
    BS_id_set.sort()
    data_processed = [np.sort(data["Binding Site {}".format(bs_id)].dropna().tolist())
                      for bs_id in BS_id_set]
    colors = [next(color_set) for dummy in BS_id_set]
    fig, ax = plt.subplots(1, 1, figsize=(len(BS_id_set)*0.6, 2.8))
    plt.subplots_adjust(bottom=0.20, top=0.83)
    ax.set_title(title, fontsize=10, weight="bold")
    parts = ax.violinplot(data_processed, showmeans=False, showmedians=False, showextrema=False)
    for pc_idx, pc in enumerate(parts['bodies']):
        pc.set_facecolors(colors[pc_idx])
        pc.set_edgecolor('black')
        pc.set_alpha(1)

    # deal with the situation in which the columns in data have different lengths.
    quartile1, medians, quartile3 = np.array([np.percentile(d, [25, 50, 75]) for d in data_processed]).T
    whiskers = np.array([
        adjacent_values(sorted_array, q1, q3)
        for sorted_array, q1, q3 in zip(data_processed, quartile1, quartile3)])
    whiskers_min, whiskers_max = whiskers[:, 0], whiskers[:, 1]

    inds = np.arange(1, len(medians) + 1)
    ax.scatter(inds, medians, marker='o', color='white', s=3, zorder=3)
    ax.vlines(inds, quartile1, quartile3, color='k', linestyle='-', lw=5)
    ax.vlines(inds, whiskers_min, whiskers_max, color='k', linestyle='-', lw=1)

    ax.get_xaxis().set_tick_params(direction='out')
    ax.xaxis.set_ticks_position('bottom')
    ax.set_xticks(np.arange(1, len(BS_id_set) + 1))
    ax.set_xticklabels(BS_id_set, fontsize=10, weight="bold")
    ax.set_xlim(0.25, len(BS_id_set) + 0.75)
    ax.set_xlabel('Binding Site', fontsize=10, weight="bold")
    ax.set_ylabel(ylabel, fontsize=10, weight="bold")
    plt.tight_layout()
    fig.savefig(fig_fn, dpi=200)
    if fig_close:
        plt.close()
    return


def plot_surface_area(surface_area, fig_fn, timeunit=None, fig_close=False):
    """Plot surface area as a function of time.

    Parameters
    ----------
    surface_area : pandas.DataFrame
    save_dir : str
    timeunit : str or None, optional, default=None

    See also
    ---------
    pylipid.func.calculate_surface_area
        The function that generates surface_area data.
    pylipid.plot.plot_surface_area_stats
        The function that plot surface_area in a matplotlib violin plot.

    """
    from itertools import cycle as _cycle

    color_set = _cycle(plt.get_cmap("tab10").colors)
    plt.rcParams["font.size"] = 10
    plt.rcParams["font.weight"] = "normal"

    if timeunit is None:
        timeunit = ""
    elif timeunit == "ns":
        timeunit = " (ns)"
    elif timeunit == "us":
        timeunit = r" ($\mu$s)"

    row_set = list(set([ind[:2] for ind in surface_area.index]))
    row_set.sort()
    col_set = [col for col in surface_area.columns if col != "Time"]
    colors = [next(color_set) for dummy in col_set]
    fig, axes = plt.subplots(len(row_set), len(col_set), figsize=(len(col_set)*2.4, len(row_set)*1.6),
                             sharex=True, sharey=True)
    plt.subplots_adjust(wspace=0.2, hspace=0.16)
    if len(col_set) == 1:
        axes = axes[:, np.newaxis]
    for row_idx, row in enumerate(row_set):
        df = surface_area.loc[row]
        for col_idx, bs_name in enumerate(col_set):
            axes[row_idx, col_idx].plot(df["Time"], df[bs_name], color=colors[col_idx],
                                        label="traj {} prot {}".format(row[0], row[1]))
            if row_idx == len(row_set)-1:
                axes[row_idx, col_idx].set_xlabel("Time{}".format(timeunit), fontsize=10)
            if col_idx == 0:
                axes[row_idx, col_idx].set_ylabel(r"Area (nm$^2$)", fontsize=10)
            if row_idx == 0:
                axes[row_idx, col_idx].set_title(bs_name, fontsize=10)
            axes[row_idx, col_idx].legend(loc="best", frameon=False)
    fig.tight_layout()
    fig.savefig(fig_fn, dpi=200)
    if fig_close:
        plt.close()

    return


class AxisIndex:
    """Build axes for logo figure."""

    def __init__(self, residue_index, logos, interactions, length, gap):
        self.page_idx = 0
        self.length = length
        self.gap = gap
        self.residue_index = residue_index
        self.logos = logos
        self.interactions = interactions
        self.axis_start = (residue_index[0] // length) * length
        self.breaks = defaultdict(list)
        self.breaks[self.page_idx].append([])
        self.gray_areas = defaultdict(list)

    def fill_missing(self, start_value, end_value):
        for xloci in np.arange(start_value, end_value + 1):
            self.breaks[self.page_idx][-1].append((xloci, "A", 0))
        self.gray_areas[self.page_idx].append((len(self.breaks[self.page_idx]) - 1, start_value, end_value))

    def new_axis(self, pointer):
        self.breaks[self.page_idx].append([])
        self.axis_start = self.residue_index[pointer]
        self.breaks[self.page_idx][-1].append(
            (self.residue_index[pointer], self.logos[pointer], self.interactions[pointer]))

    def new_page(self, pointer):
        if len(self.breaks[self.page_idx][-1]) < self.length:
            self.fill_missing(self.axis_start + len(self.breaks[self.page_idx][-1]), self.axis_start + self.length - 1)
        self.page_idx += 1
        self.breaks[self.page_idx].append([])
        self.axis_start = (self.residue_index[pointer] // self.length) * self.length
        if self.axis_start != self.residue_index[pointer]:
            self.fill_missing(self.axis_start, self.residue_index[pointer] - 1)
        self.breaks[self.page_idx][-1].append(
            (self.residue_index[pointer], self.logos[pointer], self.interactions[pointer]))

    def new_gap(self, pointer):
        gray_start = self.residue_index[pointer - 1] + 1
        for xloci in np.arange(self.residue_index[pointer - 1] + 1, self.residue_index[pointer]):
            if xloci - self.axis_start < self.length:
                self.breaks[self.page_idx][-1].append((xloci, "A", 0))
            else:
                self.gray_areas[self.page_idx].append((len(self.breaks[self.page_idx]) - 1, gray_start, xloci - 1))
                self.breaks[self.page_idx].append([])
                self.breaks[self.page_idx][-1].append((xloci, "A", 0))
                self.axis_start = xloci
                gray_start = xloci
        self.gray_areas[self.page_idx].append(
            (len(self.breaks[self.page_idx]) - 1, gray_start, self.residue_index[pointer] - 1))
        self.breaks[self.page_idx][-1].append(
            (self.residue_index[pointer], self.logos[pointer], self.interactions[pointer]))

    def sort(self):
        end = False
        if self.axis_start != self.residue_index[0]:
            self.fill_missing(self.axis_start, self.residue_index[0] - 1)
        self.breaks[self.page_idx][-1].append((self.residue_index[0], self.logos[0], self.interactions[0]))
        pointer = 1
        while not end:
            if self.residue_index[pointer] - self.residue_index[pointer - 1] == 1 and self.residue_index[
                pointer] - self.axis_start < self.length:
                self.breaks[self.page_idx][-1].append(
                    (self.residue_index[pointer], self.logos[pointer], self.interactions[pointer]))
                pointer += 1
            elif self.residue_index[pointer] - self.residue_index[pointer - 1] == 1 and self.residue_index[
                pointer] - self.axis_start >= self.length:
                self.new_axis(pointer)
                pointer += 1
            elif self.residue_index[pointer] - self.residue_index[pointer - 1] < 0:
                self.new_page(pointer)
                pointer += 1
            elif 1 < self.residue_index[pointer] - self.residue_index[pointer - 1] <= self.gap:
                self.new_gap(pointer)
                pointer += 1
            elif self.residue_index[pointer] - self.residue_index[pointer - 1] > self.gap:
                self.new_page(pointer)
                pointer += 1
            if pointer == len(self.residue_index):
                end = True
        if len(self.breaks[self.page_idx][-1]) < self.length:
            self.fill_missing(self.axis_start + len(self.breaks[self.page_idx][-1]), self.axis_start + self.length - 1)



##############################################################################
# PyLipID: A python module for analysing protein-lipid interactions
#
# Author: Wanling Song
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
##############################################################################

import numpy as np

def rmsd(V, W):
    """
    Calculate Root-mean-square deviation from two sets of vectors V and W.
    Parameters
    ----------
    V : array
        (N,D) matrix, where N is points and D is dimension.
    W : array
        (N,D) matrix, where N is points and D is dimension.
    Returns
    -------
    rmsd : scalar
        Root-mean-square deviation between the two vectors
    """
    diff = np.array(V) - np.array(W)
    N = len(V)
    return np.sqrt((diff * diff).sum() / N)



##############################################################################
# PyLipID: A python module for analysing protein-lipid interactions
#
# Author: Wanling Song
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
##############################################################################

"""This module contains functions for 2D plot.
"""

import os
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import numpy as np

__all__ = ["plot_corrcoef"]


def plot_corrcoef(corrcoef, residue_index, cmap="Reds", vmin=None, vmax=None,
                  fn=None, title=None, fig_close=False):
    """Plot correlation coefficient matrix.

    Parameters
    ----------
    corrcoef : array_like
        A scalar 2D array of correlation coefficient matrix.
    residue_index : array_list, optional, default=None
        A 1D array of residue index.
    cmap : str or `matplotlib.colors.Colormap`, optional, default="coolwarm"
        A Colormap instance or matplotlib register colormap name. The
        colormap maps *corrcoef* to colors.
    fn : str, optional, default=None
        Figure name. By default the figure is saved as "Figure_corrcoef.pdf" as the current
        working directory.
    title : str, optional, default=None

    """
    plt.rcParams["font.size"] = 10
    plt.rcParams["font.weight"] = "bold"

    if fn is None:
        fn = os.path.join(os.getcwd(), "Figure_Correlation_Matrix.pdf")

    if len(corrcoef) <= 20:
        fig, ax = plt.subplots(1, 1, figsize=(2.5, 1.8))
        majorlocator = 5
        minorlocator = 1
    elif 20 < len(corrcoef) <= 50:
        fig, ax = plt.subplots(1, 1, figsize=(2.9, 2.0))
        majorlocator = 10
        minorlocator = 1
    elif 50 < len(corrcoef) <= 500:
        fig, ax = plt.subplots(1, 1, figsize=(4.9, 3.5))
        majorlocator = 50
        minorlocator = 10
    elif 500 <= len(corrcoef) < 1000:
        fig, ax = plt.subplots(1, 1, figsize=(5.9, 4.5))
        majorlocator = 100
        minorlocator = 10
    elif 1000 <= len(corrcoef) < 2000:
        fig, ax = plt.subplots(1, 1, figsize=(7.9, 6.5))
        majorlocator = 200
        minorlocator = 20
    elif len(corrcoef) >= 2000:
        fig, ax = plt.subplots(1, 1, figsize=(8.9, 7.5))
        majorlocator = 500
        minorlocator = 100

    # sort index, check duplicates in residue index.
    majorticks = []
    minorticks = []
    ticklabels = []
    breaks = []
    for idx, resi in enumerate(residue_index):
        if not resi%majorlocator:
            majorticks.append(idx-0.5)
            ticklabels.append(resi)
        elif not resi%minorlocator:
            minorticks.append(idx-0.5)
        if idx > 0 and resi - residue_index[idx-1] != 1:
            breaks.append(idx-0.5)

    x = y = np.arange(len(residue_index)+1, dtype=float)
    x -= 0.5
    y -= 0.5

    corrcoef = np.nan_to_num(corrcoef)
    if vmax is None:
        vmax = np.percentile(np.unique(np.ravel(corrcoef)), 99)
    if vmin is None:
        vmin = np.percentile(np.unique(np.ravel(corrcoef)), 1)
    pcm = ax.pcolormesh(x, y, corrcoef, cmap=cmap, 
                        norm=colors.LogNorm(vmax=vmax, vmin=vmin))
    fig.colorbar(pcm, ax=ax)
    # set ticks
    ax.set_xticks(majorticks)
    ax.set_xticks(minorticks, minor=True)
    ax.xaxis.set_ticklabels(ticklabels, fontsize=10, weight="bold")
    ax.set_yticks(majorticks)
    ax.set_yticks(minorticks, minor=True)
    ax.yaxis.set_ticklabels(ticklabels, fontsize=10, weight="bold")
    ax.tick_params(which='both', direction='out')

    ax.set_xlabel("Residue Index", fontsize=10, weight="bold")
    ax.set_ylabel("Residue Index", fontsize=10, weight="bold")

    if len(breaks) > 0:
        for break_line in breaks:
            ax.axhline(break_line, linewidth=0.8, color="black", linestyle="--")
            ax.axvline(break_line, linewidth=0.8, color="black", linestyle="--")

    if title is not None:
        ax.set_title(title, fontsize=8, weight="bold")

    plt.tight_layout()
    fig.savefig(fn, dpi=200)
    if fig_close:
        plt.close()

    return




##############################################################################
# PyLipID: A python module for analysing protein-lipid interactions
#
# Author: Wanling Song
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
##############################################################################

"""This module contains the class Durations for using dual-cutoff scheme.
"""
import numpy as np

__all__ = ["cal_contact_residues", "Duration", "cal_occupancy", "cal_lipidcount"]


def cal_contact_residues(dist_matrix, cutoff):
    """Obtain contact residues as a function of time.

    Parameters
    ----------
    dist_matrix : list of lists or ndarray, shape=(n_residues, n_frames)
        The residue distances to the target. The distances of a residue
        to the target is listed in a row/a list as a function of time.
    cutoff : scalar
        The distance cutoff to define a contact. A distance to the target
        equal or lower to the `cutoff` is considered as in contact.

    Returns
    -------
    contact_list : list of lists
        A list of n_frame lists that contains the residues (represented by the row index of dist_matrix)
        within the cutoff distance to the target in each frame.
    frame_id_set : ndarray
        An array of frame indices where distances are smaller than the `cutoff`.
    residue_id_set : ndarray
        An array of residue indices which meet the distance

    Examples
    --------
    >>> dr0 = [0.9, 0.95, 1.2, 1.1, 1.0, 0.9] # the distances of R0 to the target as a function of time
    >>> dr1 = [0.95, 0.9, 0.95, 1.1, 1.2, 1.1] # the distances of R1
    >>> dr2 = [0.90, 0.90, 0.85, 0.95, 1.0, 1.1] # the distances of R2
    >>> dist_matrix = [dr0, dr1, dr2]
    >>> contact_list, frame_id_set, residue_id_set = cal_contact_residues(dist_matrix, 1.0)
    >>> print(contact_list)
    [[0, 1, 2], [0, 1, 2], [1, 2], [2], [0, 2], [0]]
    >>> print(frame_id_set)
    array([0, 1, 4, 5, 0, 1, 2, 0, 1, 2, 3, 4])
    >>> print(residue_id_set)
    array([0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 2])

    """
    residue_id_set, frame_id_set = np.where(np.array(dist_matrix) <= cutoff)
    contact_list = [[] for dummy in np.arange(len(dist_matrix[0]))]
    for frame_id, residue_id in zip(frame_id_set, residue_id_set):
        contact_list[frame_id].append(residue_id)
    return contact_list, frame_id_set, residue_id_set


class Duration:
    def __init__(self, contact_low, contact_high, dt):
        """Dual cutoff scheme for calculating the interaction durations.

        Parameters
        ----------
        contact_low : list of lists
            A list of n_frame lists that contains the residues within the smaller
            distance as a function of trajectory frames.
        contact_high : list of lists
            A list of n_frame lists that contains the residues within the larger
            distance as a function of trajectory frames.
        dt : scalar
            The timestep between two adjacent trajectory frames.

        """
        self.contact_low = contact_low
        self.contact_high = contact_high
        self.dt = dt
        self.pointer = [np.zeros_like(self.contact_high[idx], dtype=np.int)
                        for idx in range(len(self.contact_high))]
        return

    def cal_durations(self):
        """Calculate interaction durations using the dual-cutoff scheme.

        Calculate the durations of the appearances that start from the point when an object appears
        in `contact_low` to the point when the object disappears from `contact_high`.

        Returns
        -------
        durations : list
            A list of durations of all the interactions defined by *contact_low* and *contact_high*

        """

        durations = []
        for i in range(len(self.contact_low)):
            for j in range(len(self.contact_low[i])):
                pos = np.where(self.contact_high[i] == self.contact_low[i][j])[0][0]
                if self.pointer[i][pos] == 0:
                    durations.append(self._get_duration(i, pos))
        if len(durations) == 0:
            return [0]
        else:
            durations.sort()
            return durations

    def _get_duration(self, i, j):
        count = 1
        self.pointer[i][j] = 1
        lipid_to_search = self.contact_high[i][j]
        for k in range(i+1, len(self.contact_high)):
            locations = np.where(self.contact_high[k] == lipid_to_search)[0]
            if len(locations) == 0:
                return count * self.dt
            else:
                pos = locations[0]
                self.pointer[k][pos] = 1
                count +=1
        return (count - 1) * self.dt


def cal_occupancy(contact_list):
    """Calculate the percentage of frames in which a contact is formed.

    Parameters
    ___________
    contact_list : list of lists
        A list of lists that contains the residues (represented by the row index of dist_matrix)
        within the cutoff distance to the target in each frame.

    Returns
    -------
    Ocupancy : scalar
        The percentage of frames in which a contact is formed

    See also
    --------
    pylipid.func.cal_contact_residues
        The function that calculates contact residues from distance matrix.
    pylipid.func.cal_lipidcount
        The function that calculates the average number of contacts in all frames

    """
    if len(contact_list) == 0:
        return 0, 0
    else:
        contact_counts = [len(item) for item in contact_list]
        mask = np.array(contact_counts) > 0
        contact_counts_nonzero = np.array(contact_counts)[mask]
        return 100 * len(contact_counts_nonzero)/len(contact_list)


def cal_lipidcount(contact_list):
    """Calculate the average number of contacts at a frame.

    Parameters
    ___________
    contact_list : list of lists
        A list of lists that contains the residues (represented by the row index of dist_matrix)
        within the cutoff distance to the target in each frame.

    Returns
    -------
    LipidCount : scalar
        The average number of contacts in a frame.

    See also
    --------
    pylipid.func.cal_contact_residues
        The function that calculates contact residues from distance matrix.
    pylipid.func.cal_occupancy
        The function that calculates the percentage of frames in which a contact is formed

    """
    if len(contact_list) == 0:
        return 0, 0
    else:
        contact_counts = [len(item) for item in contact_list]
        mask = np.array(contact_counts) > 0
        contact_counts_nonzero = np.array(contact_counts)[mask]

        if len(contact_counts_nonzero) == 0:
            return 0
        else:
            return np.nan_to_num(contact_counts_nonzero.mean())




##############################################################################
# PyLipID: A python module for analysing protein-lipid interactions
#
# Author: Wanling Song
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
##############################################################################

r"""
func module
==============
The ``func`` module provides functions for functions that do the heavy-lifting:

.. currentmodule:: pylipid.func

.. autosummary::
   :toctree: generated/

   cal_koff
   cal_survival_func
   Duration
   cal_contact_residues
   cal_interaction_frequency
   get_node_list
   collect_bound_poses
   vectorize_poses
   calculate_scores
   write_bound_poses
   cluster_DBSCAN
   cluster_KMeans
   calculate_site_surface_area

"""

from .kinetics import cal_koff
from .kinetics import cal_survival_func
from .interactions import Duration
from .interactions import cal_contact_residues, cal_occupancy, cal_lipidcount
from .binding_site import get_node_list
from .binding_site import collect_bound_poses, vectorize_poses, calculate_scores, write_bound_poses
from .clusterer import cluster_DBSCAN, cluster_KMeans
from .binding_site import calculate_site_surface_area




##############################################################################
# PyLipID: A python module for analysing protein-lipid interactions
#
# Author: Wanling Song
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
##############################################################################

"""This module contains functions that deals with trajectories. """

from collections import defaultdict
import numpy as np


__all__ = ["get_traj_info"]


def get_traj_info(traj, lipid, lipid_atoms=None, resi_offset=0, nprot=1, protein_ref=None, lipid_ref=None):
    """Get trajectory information regarding atom/residue index and topologies.

    Parameters
    ----------
    traj : mdtraj.Trajectory
        A mdtraj.Trajectory object.
    lipid : str
        The residue name of the lipid to check.
    lipid_atoms : a list of str; opt
        The names of lipid atoms that are used to define lipid interaction and lipid binding sites.
        Default is None, that is all the lipid atoms will be used for calculation.
    resi_offset : int, optional, default=0
        Shift of residue index. The new residue index (i.e. the original index + resi_offset) will
        be used in all the generated data.
    nprot : int, optional, default=1
        Number of protein copies in the systems. If nprot >= 2, the protein copies need to be identical,
        and the generated data will be the averages of the copies.
    protein_ref : None or mdtraj.Trajectory, optional, default=None
        A mdtraj.Trajectory object that stores the topology and coordinates of a copy of the protein structure.
    lipid_ref : None or mdtraj.Trajectory, optional, default=None
        A mdtraj.Trajectory object that stores the topology and coordinates of a lipid molecule structure.

    Returns
    -------
    traj_info : dict
        A dictionary that contains the topology information of `traj`.
    protein_ref :  mdtraj.Trajectory
        A mdtraj.Trajectory object that stores the topology and coordinates of a copy of the protein structure.
    lipid_ref : mdtraj.Trajectory
        A mdtraj.Trajectory object that stores the topology and coordinates of a lipid molecule structure.

    """
    lipid_index_dict = defaultdict(list)
    lipid_atom_indices = traj.top.select("resn {}".format(lipid))
    if lipid_atoms is not None:
        for atom_index in lipid_atom_indices:
            if traj.top.atom(atom_index).name in lipid_atoms:
                lipid_index_dict[traj.top.atom(atom_index).residue.index].append(atom_index)
    else:
        for atom_index in lipid_atom_indices:
            lipid_index_dict[traj.top.atom(atom_index).residue.index].append(atom_index)
    lipid_residue_atomid_list = [lipid_index_dict[resi] for resi in np.sort(list(lipid_index_dict.keys()))]
    # get protein atom indices
    all_protein_atom_indices = traj.top.select("protein")
    natoms_per_protein = int(len(all_protein_atom_indices)/nprot)
    protein_residue_atomid_list = []
    for protein_idx in np.arange(nprot):
        chain_index_dict = defaultdict(list)
        for atom_index in all_protein_atom_indices[protein_idx*natoms_per_protein:(protein_idx+1)*natoms_per_protein]:
            chain_index_dict[traj.top.atom(atom_index).residue.index].append(atom_index)
        protein_residue_atomid_list.append([chain_index_dict[resi] for resi in np.sort(list(chain_index_dict.keys()))])
    protein_residue_id = np.arange(len(protein_residue_atomid_list[0]), dtype=int)
    residue_list = np.array(["{}{}".format(traj.top.atom(residue_atoms[0]).residue.resSeq+resi_offset,
                                           traj.top.atom(residue_atoms[0]).residue.name)
                             for residue_atoms in protein_residue_atomid_list[0]], dtype=str)
    if lipid_ref is None:
        one_lipid_indices = []
        for lipid_id in np.sort(traj.top.select("resn {}".format(lipid))):
            if len(one_lipid_indices) == 0:
                one_lipid_indices.append(lipid_id)
            elif traj.top.atom(lipid_id).residue.index != traj.top.atom(one_lipid_indices[-1]).residue.index:
                break
            else:
                one_lipid_indices.append(lipid_id)
        lipid_ref = traj[0].atom_slice(np.unique(one_lipid_indices))
    if protein_ref is None:
        protein_ref = traj[0].atom_slice(all_protein_atom_indices[:natoms_per_protein])

    traj_info = {"protein_residue_atomid_list": protein_residue_atomid_list,
                 "lipid_residue_atomid_list": lipid_residue_atomid_list,
                 "protein_residue_id": protein_residue_id, "residue_list": residue_list}

    return traj_info, protein_ref, lipid_ref



##############################################################################
# PyLipID: A python module for analysing protein-lipid interactions
#
# Author: Wanling Song
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
##############################################################################

r"""
PyLipID class
=============
**api** provides the outer layer functions to calculate lipid interactions
and binding sites.

.. currentmodule:: pylipid.api

.. autosummary::
   :toctree: generated/

   LipidInteraction

"""

from .api import LipidInteraction


##############################################################################
# PyLipID: A python module for analysing protein-lipid interactions
#
# Author: Wanling Song
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
##############################################################################

"""This module contains functions that write PyMol scripts."""

import os
from shutil import copyfile

__all__ = ["write_pymol_script"]


def write_pymol_script(fname, pdb_fname, data_fname, lipid, n_binding_site):
    """Write PyMol script.

    Parameters
    ----------
    fname : str
    pdb_fname : str
    data_fname : str
    lipid : str
    n_binding_site : int

    """
    script_fname = os.path.abspath(fname)
    save_dir, _ = os.path.split(script_fname)
    pdb_new_fname = os.path.join(save_dir, os.path.basename(pdb_fname))
    copyfile(os.path.abspath(pdb_fname), pdb_new_fname)
    data_fname = os.path.abspath(data_fname)

    text = """
import numpy as np
import re
import pymol
from pymol import cmd
pymol.finish_launching()

########## files to process ##########
csv_file = "{CSV_FILE}"
pdb_file = "{PDB}"

########## set the sphere scales to corresponding value ##########
value_to_show = "Residence Time"

###### reading data from csv file ##########
num_of_binding_site = {N_BINDING_SITE}

with open(csv_file, "r") as f:
    data_lines = f.readlines()

column_names = data_lines[0].strip().split(",")
for column_idx, column_name in enumerate(column_names):
    if column_name == "Residue":
        column_id_residue_list = column_idx
    elif column_name == "Residue ID":
        column_id_residue_index = column_idx
    elif column_name == "Binding Site ID":
        column_id_BS = column_idx
    elif column_name == value_to_show:
        column_id_value_to_show = column_idx

residue_list = []
residue_rank_set = []
binding_site_identifiers = []
values_to_show = []
for line in data_lines[1:]:
    data_list = line.strip().split(",")
    residue_list.append(data_list[column_id_residue_list])
    residue_rank_set.append(data_list[column_id_residue_index])
    binding_site_identifiers.append(float(data_list[column_id_BS]))
    values_to_show.append(data_list[column_id_value_to_show])

############## read information from pdb file ##########
with open(pdb_file, "r") as f:
    pdb_lines = f.readlines()
residue_identifiers = []
for line in pdb_lines:
    line_stripped = line.strip()
    if line_stripped[:4] == "ATOM":
        identifier = (line_stripped[22:26].strip(), line_stripped[17:20].strip(), line_stripped[21].strip())
##                           residue index,              resname,                     chain id
        if len(residue_identifiers) == 0:
            residue_identifiers.append(identifier)
        elif identifier != residue_identifiers[-1]:
            residue_identifiers.append(identifier)

######### set sphere scales ###############
values_to_show = np.array(values_to_show, dtype=float)
MIN = np.percentile(np.unique(values_to_show), 5)
MAX = np.percentile(np.unique(values_to_show), 100)
X = (values_to_show - np.percentile(np.unique(values_to_show), 50))/(MAX - MIN)
SCALES = 1/(0.5 + np.exp(-X * 5))

######## some pymol settings #########
cmd.set("retain_order", 1)
cmd.set("cartoon_oval_length", 1.0)
cmd.set("cartoon_oval_width", 0.3)
cmd.set("cartoon_color", "white")
cmd.set("stick_radius", 0.35)

##################################
cmd.load(pdb_file, "Prot_{LIPID}")
prefix = "Prot_{LIPID}"
cmd.hide("everything")
cmd.show("cartoon", prefix)
cmd.center(prefix)
cmd.orient(prefix)
colors = np.array([np.random.choice(np.arange(256, dtype=float), size=3) for dummy in range(num_of_binding_site)])
colors /= 255.0
            """.format(**{"SAVE_DIR": save_dir, "LIPID": lipid, "N_BINDING_SITE": n_binding_site,
                          "PDB": pdb_new_fname, "CSV_FILE": data_fname})

    text += r"""
residue_list = np.array(residue_list, dtype=str)
residue_rank_set = np.array(residue_rank_set, dtype=int)
binding_site_identifiers = np.array(binding_site_identifiers, dtype=int)
residue_identifiers = list(residue_identifiers)
for bs_id in np.arange(num_of_binding_site):
    cmd.set_color("tmp_{}".format(bs_id), list(colors[bs_id]))
    for entry_id in np.where(binding_site_identifiers == bs_id)[0]:
        selected_residue = residue_list[entry_id]
        selected_residue_rank = residue_rank_set[entry_id]
        identifier_from_pdb = residue_identifiers[selected_residue_rank]
        if re.findall("[a-zA-Z]+$", selected_residue)[0] != identifier_from_pdb[1]:
            raise IndexError(
            "The {}th residue in the provided pdb file ({}{}) is different from that in the simulations ({})!".format(
                                                                                            entry_id+1,
                                                                                            identifier_from_pdb[0],
                                                                                            identifier_from_pdb[1],
                                                                                            selected_residue)
                                                                                            )
        if identifier_from_pdb[2] != " ":
            cmd.select("BSid{}_{}".format(bs_id, selected_residue), 
            "chain {} and resid {} and (not name C+O+N)".format(identifier_from_pdb[2], identifier_from_pdb[0]))
        else:
            cmd.select("BSid{}_{}".format(bs_id, selected_residue), 
            "resid {} and (not name C+O+N)".format(identifier_from_pdb[0]))
        cmd.show("spheres", "BSid{}_{}".format(bs_id, selected_residue))
        cmd.set("sphere_scale", SCALES[entry_id], selection="BSid{}_{}".format(bs_id, selected_residue))
        cmd.color("tmp_{}".format(bs_id), "BSid{}_{}".format(bs_id, selected_residue))
    cmd.group("BSid{}".format(bs_id), "BSid{}_*".format(bs_id))

            """
    with open(fname, "w") as f:
        f.write(text)

    return


##############################################################################
# PyLipID: A python module for analysing protein-lipid interactions
#
# Author: Wanling Song
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
##############################################################################

from collections import defaultdict
from itertools import product
import pickle
import os
import re
import warnings
import mdtraj as md
import numpy as np
np.seterr(all='ignore')
from scipy.sparse import coo_matrix
from sklearn.decomposition import PCA
import pandas as pd
from tqdm import trange, tqdm
from ..func import cal_contact_residues
from ..func import Duration, cal_koff
from ..func import cal_lipidcount, cal_occupancy
from ..func import get_node_list
from ..func import collect_bound_poses, vectorize_poses, calculate_scores, write_bound_poses
from ..func import cluster_DBSCAN, cluster_KMeans
from ..func import calculate_site_surface_area
from ..plot import plot_koff
from ..plot import plot_surface_area, plot_binding_site_data
from ..plot import plot_residue_data, plot_corrcoef, plot_residue_data_logos
from ..util import check_dir, write_PDB, write_pymol_script, sparse_corrcoef, rmsd, get_traj_info


class LipidInteraction:
    def __init__(self, trajfile_list, cutoffs=[0.475, 0.7], lipid="CHOL", topfile_list=None, lipid_atoms=None,
                 nprot=1, resi_offset=0, save_dir=None, timeunit="us", stride=1, dt_traj=None):

        """The outer layer class that integrates calculations and handles workflow.

        `LipidInteraction` reads trajectory information via `mdtraj.load()`, and calculate interactions
        of the specified lipid for both protein residues and the calculated binding sites. 'LipidInteraction' also
        has a couple of assisting functions to plot the interaction data and save the generated data.

        Parameters
        ----------
        trajfile_list : str or a list of str
            Trajectory filename(s). Read by mdtraj.load() to obtain trajectory information.

        cutoffs : list of two scalar or a scalar, default=[0.475, 0.7]
            Cutoff value(s) for defining contacts. When a list of two scalar are provided, the dual-cutoff scheme
            will be used.

        lipid : str, default="CHOL"
            Lipid name in topology.

        topfile_list : str or a list of str, default=None
            Topology filename(s). Most trajectory formats do not contain topology information. Provide either
            the path to a RCSB PDB file, a trajectory, or a topology for each trajectory in `trajfile_list`
            for the topology information. See `mdtraj.load() <https://mdtraj.org>`_. for more information.

        lipid_atoms : list of str, default=None
            Lipid atom names. Only interactions of the provided atoms will be considered for the calculation of contacts.
            If None, all atoms of the lipid molecule will be used.

        nprot : int, default=1
            Number of protein copies in the system. If the system has N copies of the protein, 'nprot=N' will report
            averaged values from the N copies, but 'nprot=1' will report interaction values for each copy.

        resi_offset : int, default=0
            Shift residue index in the reported results from what is shown in the topology. Can be useful for
            MARTINI force field.

        save_dir : str, default=None
            The root directory to store the data. All the generated dataset and created directories will be
            put under this directory.

        timeunit : {"us", "ns"}, default="us"
            The time unit used for reporting results. "us" is micro-second and "ns" is nanosecond.

        stride : int, default=1
            Only read every stride-th frame. The same stride in mdtraj.load().

        dt_traj : float, default=None
            Timestep of trajectories. It is required when trajectories do not have timestep information. Not needed for
            trajectory formats of e.g. xtc, trr etc. If None, timestep information will take from trajectories.

        """
        self._trajfile_list = np.atleast_1d(trajfile_list)
        if len(np.atleast_1d(topfile_list)) == len(self._trajfile_list):
            self._topfile_list = np.atleast_1d(topfile_list)
        elif len(self._trajfile_list) > 1 and len(np.atleast_1d(topfile_list)) == 1:
            self._topfile_list = [topfile_list for dummy in self._trajfile_list]
        else:
            raise ValueError(
                "topfile_list should either have the same length as trajfile_list or have one valid file name.")

        if len(np.atleast_1d(cutoffs)) == 1:
            self._cutoffs = np.array([np.atleast_1d(cutoffs)[0] for dummy in range(2)])
        elif len(np.atleast_1d(cutoffs)) == 2:
            self._cutoffs = np.sort(np.array(cutoffs, dtype=float))
        else:
            raise ValueError("cutoffs should be either a scalar or a list of two scalars.")

        self._dt_traj = dt_traj
        self._lipid = lipid
        self._lipid_atoms = lipid_atoms
        self._nprot = int(nprot)
        self._timeunit = timeunit
        self._stride = int(stride)
        self._resi_offset = resi_offset
        self.dataset = pd.DataFrame()
        self._save_dir = check_dir(os.getcwd(), "Interaction_{}".format(self._lipid)) if save_dir is None \
            else check_dir(save_dir, "Interaction_{}".format(self._lipid))
        return

    #############################################
    #     attributes
    #############################################
    @property
    def residue_list(self):
        """Residue names."""
        return self._residue_list

    @property
    def node_list(self):
        """Residue ID list of binding site"""
        return self._node_list

    @property
    def lipid(self):
        """Lipid residue name."""
        return self._lipid

    @property
    def lipid_atoms(self):
        """Lipid atom names"""
        return self._lipid_atoms

    @property
    def cutoffs(self):
        """Cutoffs used for calculating contacts. """
        return self._cutoffs

    @property
    def nprot(self):
        """Number of protein copies in system. """
        return self._nprot

    @property
    def stride(self):
        """Stride"""
        return self._stride

    @property
    def trajfile_list(self):
        """Trajectory filenames """
        return self._trajfile_list

    @property
    def topfile_list(self):
        """Topology filenames"""
        return self._topfile_list

    @property
    def dt_traj(self):
        """Trajectory timestep"""
        return self._dt_traj

    @property
    def resi_offset(self):
        """Residue index offset"""
        return self._resi_offset

    @property
    def save_dir(self):
        """Root directory of the generated data."""
        return self._save_dir

    @property
    def timeunit(self):
        """Time unit used for reporting results. """
        return self._timeunit

    def koff(self, residue_id=None, residue_name=None):
        """Residue koff"""
        if residue_id is not None and residue_name is not None:
            assert self.dataset[self.dataset["Residue ID"] == residue_id]["Residue"] == residue_name, \
                "residue_id and residue_name are pointing to different residues!"
            return self._koff[residue_id]
        elif residue_id is not None:
            return self._koff[residue_id]
        elif residue_name is not None:
            return self._koff[self._residue_map[residue_name]]

    def res_time(self, residue_id=None, residue_name=None):
        """Residue residence time"""
        if residue_id is not None and residue_name is not None:
            assert self.dataset[self.dataset["Residue ID"] == residue_id]["Residue"] == residue_name, \
                "residue_id and residue_name are pointing to different residues!"
            return self._res_time[residue_id]
        elif residue_id is not None:
            return self._res_time[residue_id]
        elif residue_name is not None:
            return self._res_time[self._residue_map[residue_name]]

    def koff_bs(self, bs_id):
        """Binding site koff"""
        return self._koff_BS[bs_id]

    def res_time_bs(self, bs_id):
        """Binding site residence time"""
        return self._res_time_BS[bs_id]

    def residue(self, residue_id=None, residue_name=None, print_data=True):
        """Obtain the lipid interaction information for a residue

        Use either residue_id or residue_name to obtain the information.
        Return the information in a pandas.DataFrame object.

        Parameters
        ----------
        residue_id : int or list of int, default=None
            The residue ID that is used by PyLipID for identifying residues. The ID starts from 0, i.e. the ID
            of N-th residue is (N-1). If None, all residues are selected.
        residue_name : str or list of str, default=None
            The residue name as stored in PyLipID dataset. The residue name is in the format of resi+resn

        Returns
        -------
        df : pandas.DataFrame
            A pandas.DataFrame of interaction information of the residue.

        """
        if residue_id is not None and residue_name is not None:
            assert self.dataset[self.dataset["Residue ID"] == residue_id]["Residue"] == residue_name, \
                "residue_id and residue_name are pointing to different residues!"
            df = self.dataset[self.dataset["Residue ID"] == residue_id]
        elif residue_id is not None:
            df = self.dataset[self.dataset["Residue ID"] == residue_id]
        elif residue_name is not None:
            df = self.dataset[self.dataset["Residue"] == residue_name]
        if print_data:
            print(df)
        return df

    def binding_site(self, binding_site_id, print_data=True, sort_residue="Residence Time"):
        """Obtain the lipid interaction information for a binding site.

        Use binding site ID to access the information. Return the lipid interaction information of the
        binding site in a pandas.DataFrame object. If print_data is True, the binding site info will be
        formatted and print out.

        """
        df = self.dataset[self.dataset["Binding Site ID"] == binding_site_id].sort_values(by="Residence Time")
        if print_data:
            text = self._format_BS_print_info(binding_site_id, self._node_list[binding_site_id], sort_residue)
            print(text)
        return df

    ########################################
    #     interaction calculation
    ########################################
    def collect_residue_contacts(self):
        """Create contacting lipid index for residues at each frame.

        This function needs to run before any of the calculations.

        """
        self._protein_ref = None
        self._lipid_ref = None
        self._T_total = []
        self._timesteps = []
        self._protein_residue_id = []
        # initialise data for interaction matrix
        col = []
        row = []
        data = []
        ncol_start = 0
        # calculate interactions from trajectories
        for traj_idx in trange(len(self._trajfile_list), desc="COLLECT INTERACTIONS FROM TRAJECTORIES",
                               total=len(self._trajfile_list)):
            traj = md.load(self._trajfile_list[traj_idx], top=self._topfile_list[traj_idx], stride=self._stride)
            traj_info, self._protein_ref, self._lipid_ref = get_traj_info(traj, self._lipid,
                                                                          lipid_atoms=self._lipid_atoms,
                                                                          resi_offset=self._resi_offset,
                                                                          nprot=self._nprot,
                                                                          protein_ref=self._protein_ref,
                                                                          lipid_ref=self._lipid_ref)
            if self._dt_traj is None:
                timestep = traj.timestep / 1000000.0 if self._timeunit == "us" else traj.timestep / 1000.0
            else:
                timestep = float(self._dt_traj * self._stride)
            self._T_total.append((traj.n_frames - 1) * timestep)
            self._timesteps.append(timestep)
            if len(self._protein_residue_id) == 0:
                self._protein_residue_id = traj_info["protein_residue_id"]
                self._residue_list = traj_info["residue_list"]
                self._nresi_per_protein = len(self._residue_list)
                self._duration = {residue_id: [] for residue_id in self._protein_residue_id}
                self._occupancy = {residue_id: [] for residue_id in self._protein_residue_id}
                self._lipid_count = {residue_id: [] for residue_id in self._protein_residue_id}
                self._contact_residues_high = {residue_id: [] for residue_id in self._protein_residue_id}
                self._contact_residues_low = {residue_id: [] for residue_id in self._protein_residue_id}
                self._koff = np.zeros(self._nresi_per_protein)
                self._koff_boot = np.zeros(self._nresi_per_protein)
                self._r_squared = np.zeros(self._nresi_per_protein)
                self._r_squared_boot = np.zeros(self._nresi_per_protein)
                self._res_time = np.zeros(self._nresi_per_protein)
                self._residue_map = {residue_name: residue_id
                                     for residue_id, residue_name in zip(self._protein_residue_id, self._residue_list)}
            else:
                assert len(self._protein_residue_id) == len(traj_info["protein_residue_id"]), \
                    "Trajectory {} contains {} residues whereas trajectory {} contains {} residues".format(
                        traj_idx, len(traj_info["protein_residue_id"]), traj_idx - 1, len(self._protein_residue_id))
            ncol_per_protein = len(traj_info["lipid_residue_atomid_list"]) * traj.n_frames
            for protein_idx in np.arange(self._nprot, dtype=int):
                for residue_id, residue_atom_indices in enumerate(
                        traj_info["protein_residue_atomid_list"][protein_idx]):
                    # calculate interaction per residue
                    dist_matrix = np.array([np.min(
                        md.compute_distances(traj, np.array(list(product(residue_atom_indices, lipid_atom_indices)))),
                        axis=1) for lipid_atom_indices in traj_info["lipid_residue_atomid_list"]])
                    contact_low, frame_id_set_low, lipid_id_set_low = cal_contact_residues(dist_matrix, self._cutoffs[0])
                    contact_high, _, _ = cal_contact_residues(dist_matrix, self._cutoffs[1])
                    self._contact_residues_high[residue_id].append(contact_high)
                    self._contact_residues_low[residue_id].append(contact_low)
                    # update coordinates for coo_matrix
                    col.append([ncol_start + ncol_per_protein * protein_idx + lipid_id * traj.n_frames +
                                frame_id for frame_id, lipid_id in zip(frame_id_set_low, lipid_id_set_low)])
                    row.append([residue_id for dummy in np.arange(len(frame_id_set_low), dtype=int)])
                    data.append(dist_matrix[lipid_id_set_low, frame_id_set_low])
            ncol_start += ncol_per_protein * self._nprot
        # calculate correlation coefficient matrix
        row = np.concatenate(row)
        col = np.concatenate(col)
        data = np.concatenate(data)
        contact_info = coo_matrix((data, (row, col)), shape=(self._nresi_per_protein, ncol_start))
        self.interaction_corrcoef = sparse_corrcoef(contact_info)
        self.dataset = pd.DataFrame({"Residue": [residue for residue in self._residue_list],
                                     "Residue ID": self._protein_residue_id})
        return

    def compute_residue_duration(self, residue_id=None):
        """Calculate the durations of lipid contacts for residues

        Parameters
        ----------
        residue_id : int or list of int, default=None

        Returns
        -------
        durations : list, len(durations)=len(residue_id)

        """
        self._check_calculation("Residue", self.collect_residue_contacts)
        if residue_id is None:
            selected_residue_id = self._protein_residue_id
        else:
            selected_residue_id = np.atleast_1d(residue_id)

        for residue_id in selected_residue_id:
            self._duration[residue_id] = [
                        Duration(self._contact_residues_low[residue_id][traj_idx*protein_idx],
                                 self._contact_residues_high[residue_id][traj_idx*protein_idx],
                                 self._timesteps[traj_idx]).cal_durations()
                        for traj_idx in np.arange(len(self.trajfile_list))
                        for protein_idx in np.arange(self._nprot, dtype=int)]
        self.dataset["Duration"] = [np.mean(np.concatenate(self._duration[residue_id]))
                                    if len(self._duration[residue_id]) > 0 else 0
                                    for residue_id in self._protein_residue_id]
        self.dataset["Duration std"] = [np.std(np.concatenate(self._duration[residue_id]))
                                        if len(self._duration[residue_id]) > 0 else 0
                                        for residue_id in self._protein_residue_id]

        if len(selected_residue_id) == 1:
            return self._duration[residue_id]
        else:
            return [self._duration[residue_id] for residue_id in selected_residue_id]

    def compute_residue_occupancy(self, residue_id=None):
        """Calculate the percentage of frames in which lipid contacts are formed for residues.

        Parameters
        ----------
        residue_id : int or list of int, default=None

        Returns
        -------
        occupancies : list, len(occupancies)=len(residue_id)

        """
        self._check_calculation("Residue", self.collect_residue_contacts)
        if residue_id is None:
            selected_residue_id = self._protein_residue_id
        else:
            selected_residue_id = np.atleast_1d(residue_id)
        for residue_id in selected_residue_id:
            self._occupancy[residue_id] = [cal_occupancy(self._contact_residues_low[residue_id][traj_idx*protein_idx])
                                          for traj_idx in np.arange(len(self.trajfile_list))
                                          for protein_idx in np.arange(self._nprot, dtype=int)]
        self.dataset["Occupancy"] = [np.mean(self._occupancy[residue_id])
                                     if len(self._occupancy[residue_id]) > 0 else 0
                                     for residue_id in self._protein_residue_id]
        self.dataset["Occupancy std"] = [np.std(self._occupancy[residue_id])
                                         if len(self._occupancy[residue_id]) > 0 else 0
                                         for residue_id in self._protein_residue_id]

        if len(selected_residue_id) == 1:
            return self._occupancy[residue_id]
        else:
            return [self._occupancy[residue_id] for residue_id in selected_residue_id]

    def compute_residue_lipidcount(self, residue_id=None):
        """Calculate the average number of surrounding lipids for residues.

        Parameters
        ----------
        residue_id : int or list of int, default=None

        Returns
        -------
        lipidcounts : list, len(lipidcount)=len(residue_id)

        """
        self._check_calculation("Residue", self.collect_residue_contacts)
        if residue_id is None:
            selected_residue_id = self._protein_residue_id
        else:
            selected_residue_id = np.atleast_1d(residue_id)
        for residue_id in selected_residue_id:
            self._lipid_count[residue_id] = [cal_lipidcount(self._contact_residues_low[residue_id][traj_idx * protein_idx])
                                           for traj_idx in np.arange(len(self.trajfile_list))
                                           for protein_idx in np.arange(self._nprot, dtype=int)]
            self.dataset["Lipid Count"] = [np.mean(self._lipid_count[residue_id])
                                           if len(self._lipid_count[residue_id]) > 0 else 0
                                           for residue_id in self._protein_residue_id]
            self.dataset["Lipid Count std"] = [np.std(self._lipid_count[residue_id])
                                               if len(self._lipid_count[residue_id]) > 0 else 0
                                               for residue_id in self._protein_residue_id]
        if len(selected_residue_id) == 1:
            self._lipid_count[residue_id]
        else:
            return [self._lipid_count[residue_id] for residue_id in selected_residue_id]

    def compute_residue_koff(self, residue_id=None, nbootstrap=10, initial_guess=[1., 1., 1., 1.],
                             save_dir=None, plot_data=True, fig_close=True):
        """Calculate interaction koff and residence time for residues.

        Parameters
        ----------
        residue_id : int or array_like or None, default=None
        nbootstrap : int, default=10
        initial_guess : array_like, default=None
        save_dir : str, default=None
        print_data : bool, default=True
        plot_data : bool, default=True
        fig_close : bool, default=True

        Returns
        ---------
        koff : scalar or list of scalar
        restime : scalar or list of scalar

        """
        self._check_calculation("Residue", self.compute_residue_koff)
        if plot_data:
            koff_dir = check_dir(save_dir, "Reisidue_koffs_{}".format(self._lipid)) if save_dir is not None \
                else check_dir(self._save_dir, "Residue_koffs_{}".format(self._lipid))
        self._check_calculation("Residue", self.collect_residue_contacts)
        if len(set(self._residue_list)) != len(self._residue_list):
            residue_name_set = ["{}_ResidueID{}".format(residue, residue_id) for residue, residue_id in
                                zip(self._residue_list, self._protein_residue_id)]
        else:
            residue_name_set = self._residue_list
        if residue_id is not None:
            selected_residue_id = np.atleast_1d(residue_id)
        else:
            selected_residue_id = self._protein_residue_id
        residues_missing_durations = [residue_id for residue_id in selected_residue_id
                                      if len(self._duration[residue_id]) == 0]
        if len(residues_missing_durations) > 0:
            self.compute_residue_duration(residue_id=residues_missing_durations)
        t_total = np.max(self._T_total)
        same_length = np.all(np.array(self._T_total) == t_total)
        if not same_length:
            warnings.warn(
                "Trajectories have different lengths. This will impair the accuracy of koff calculation!")
        timestep = np.min(self._timesteps)
        same_timestep = np.all(np.array(self._timesteps) == timestep)
        if not same_timestep:
            warnings.warn(
                "Trajectories have different timesteps. This will impair the accuracy of koff calculation!")

        for residue_id in tqdm(selected_residue_id,desc="CALCULATE KOFF FOR RESIDUES", total=len(selected_residue_id)):
            durations = np.concatenate(self._duration[residue_id])
            residue = residue_name_set[residue_id]
            if np.sum(durations) == 0:
                self._koff[residue_id] = 0
                self._res_time[residue_id] = 0
                self._r_squared[residue_id] = 0
                self._koff_boot[residue_id] = 0
                self._r_squared_boot[residue_id] = 0
            else:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    koff, restime, properties = cal_koff(durations, t_total, timestep, nbootstrap, initial_guess)
                self._koff[residue_id] = koff
                self._res_time[residue_id] = restime
                self._r_squared[residue_id] = properties["r_squared"]
                self._koff_boot[residue_id] = np.mean(properties["ks_boot_set"], axis=0)[0]
                self._r_squared_boot[residue_id] = np.mean(properties["r_squared_boot_set"])
                if plot_data:
                    text = self._format_koff_text(properties)
                    plot_koff(durations, properties["delta_t_list"], properties["survival_rates"],
                              properties["n_fitted"], survival_rates_bootstraps=properties["survival_rates_boot_set"],
                              fig_fn=os.path.join(koff_dir, "{}.pdf".format(residue)), title=residue,
                              timeunit=self._timeunit, t_total=t_total, text=text, fig_close=fig_close)
        # update dataset
        self.dataset["Koff"] = self._koff
        self.dataset["Residence Time"] = self._res_time
        self.dataset["R Squared"] = self._r_squared
        self.dataset["Koff Bootstrap avg"] = self._koff_boot
        self.dataset["R Squared Bootstrap avg"] = self._r_squared_boot

        if len(selected_residue_id) == 1:
            return self._koff[selected_residue_id[0]], self._res_time[selected_residue_id[0]]
        else:
            return [self._koff[residue_id] for residue_id in selected_residue_id], \
                   [self._res_time[residue_id] for residue_id in selected_residue_id]

    def compute_binding_nodes(self, threshold=4, print_data=True):
        """Calculate binding sites via computing the community structures.

        Parameters
        ----------
        threshold : int, default=4
        save_dir : str or None, default=None
        print : bool, default=True

        Returns
        -------
        node_list: list
        modularity : float or None

        """
        self._check_calculation("Residue", self.compute_residue_koff)
        corrcoef_raw = np.nan_to_num(self.interaction_corrcoef)
        corrcoef = np.copy(corrcoef_raw)
        node_list, modularity = get_node_list(corrcoef, threshold=threshold)
        self._node_list = node_list
        self._network_modularity = modularity
        if len(self._node_list) == 0:
            print("*"*30)
            print(" No binding site detected!!")
            print("*"*30)
        else:
            residue_BS_identifiers = np.ones(self._nresi_per_protein, dtype=int) * -1
            for bs_id, nodes in enumerate(self._node_list):
                residue_BS_identifiers[nodes] = int(bs_id)
            # update dataset
            self.dataset["Binding Site ID"] = residue_BS_identifiers
            # initialise variable for binding site interactions
            self._duration_BS = {bs_id:[] for bs_id in np.arange(len(self._node_list))}
            self._occupancy_BS = {bs_id:[] for bs_id in np.arange(len(self._node_list))}
            self._lipid_count_BS = {bs_id:[] for bs_id in np.arange(len(self._node_list))}
            self._koff_BS = np.zeros(len(self._node_list))
            self._koff_BS_boot = np.zeros(len(self._node_list))
            self._res_time_BS = np.zeros(len(self._node_list))
            self._r_squared_BS = np.zeros(len(self._node_list))
            self._r_squared_BS_boot = np.zeros(len(self._node_list))
            if print_data:
                print(f"Network modularity: {modularity:.3f}")
                for bs_id, nodes in enumerate(self._node_list):
                    print("#" * 25)
                    print(f"Binding Site ID: {bs_id}")
                    print("{:>10s} -- {:<12s}".format("Residue", "Residue ID"))
                    for node in nodes:
                        print("{:>10s} -- {:<12d}".format(self._residue_list[node], self._protein_residue_id[node]))
                    print("#" * 25)
        return node_list, modularity

    def compute_site_duration(self, binding_site_id=None):
        """Calculate interaction durations for binding sites.

        Parameters
        ----------
        binding_site_id : int of list of int, default=None

        Returns
        -------
        durations_BS : list, len(durations_BS)=len(binding_site_id)

        """
        self._check_calculation("Residue", self.collect_residue_contacts)
        self._check_calculation("Binding Site ID", self.compute_binding_nodes, print_data=False)
        selected_bs_id = np.atleast_1d(binding_site_id) if binding_site_id is not None \
            else np.arange(len(self._node_list), dtype=int)
        for bs_id in selected_bs_id:
            nodes = self._node_list[bs_id]
            durations_BS = []
            for traj_idx in np.arange(len(self._trajfile_list), dtype=int):
                for protein_idx in np.arange(self._nprot, dtype=int):
                    list_to_take = traj_idx * self._nprot + protein_idx
                    n_frames = len(self._contact_residues_low[nodes[0]][list_to_take])
                    contact_BS_low = [np.unique(np.concatenate(
                        [self._contact_residues_low[node][list_to_take][frame_idx] for node in nodes]))
                        for frame_idx in np.arange(n_frames)]
                    contact_BS_high = [np.unique(np.concatenate(
                        [self._contact_residues_high[node][list_to_take][frame_idx] for node in nodes]))
                        for frame_idx in np.arange(n_frames)]
                    durations_BS.append(
                        Duration(contact_BS_low, contact_BS_high, self._timesteps[traj_idx]).cal_durations())
            self._duration_BS[bs_id] = durations_BS
        # update dataset
        durations_BS_per_residue = np.zeros(self._nresi_per_protein)
        for bs_id, nodes in enumerate(self._node_list):
            durations_BS_per_residue[nodes] = np.mean(np.concatenate(self._duration_BS[bs_id])) \
                if len(self._duration_BS[bs_id]) > 0 else 0
        self.dataset["Binding Site Duration"] = durations_BS_per_residue

        if len(selected_bs_id) == 1:
            return self._duration_BS[bs_id]
        else:
            return [self._duration_BS[bs_id] for bs_id in selected_bs_id]

    def compute_site_occupancy(self, binding_site_id=None):
        """Calculate the percentage of frames in which lipid contacts are formed for binding sites.

        Parameters
        ----------
        binding_site_id : int or list of int, default=None

        Returns
        -------
        occupancy_BS : list, len(occupancy_BS)=len(binding_site_id)

        """
        self._check_calculation("Residue", self.collect_residue_contacts)
        self._check_calculation("Binding Site ID", self.compute_binding_nodes, print_data=False)
        selected_bs_id = np.atleast_1d(binding_site_id) if binding_site_id is not None \
            else np.arange(len(self._node_list), dtype=int)
        for bs_id in selected_bs_id:
            nodes = self._node_list[bs_id]
            occupancy_BS = []
            for traj_idx in np.arange(len(self._trajfile_list), dtype=int):
                for protein_idx in np.arange(self._nprot, dtype=int):
                    list_to_take = traj_idx * self._nprot + protein_idx
                    n_frames = len(self._contact_residues_low[nodes[0]][list_to_take])
                    contact_BS_low = [np.unique(np.concatenate(
                        [self._contact_residues_low[node][list_to_take][frame_idx] for node in nodes]))
                        for frame_idx in np.arange(n_frames)]
                    occupancy_BS.append(cal_occupancy(contact_BS_low))
            self._occupancy_BS[bs_id] = occupancy_BS
        # update dataset
        occupancy_BS_per_residue = np.zeros(self._nresi_per_protein)
        for bs_id, nodes in enumerate(self._node_list):
            occupancy_BS_per_residue[nodes] = np.mean(self._occupancy_BS[bs_id]) \
                if len(self._occupancy_BS[bs_id]) > 0 else 0
        self.dataset["Binding Site Occupancy"] = occupancy_BS_per_residue

        if len(selected_bs_id) == 1:
            return self._occupancy_BS[bs_id]
        else:
            return [self._occupancy_BS[bs_id] for bs_id in selected_bs_id]

    def compute_site_lipidcount(self, binding_site_id=None):
        """Calculate the average number of surrounding lipids for binding sites.

        Parameters
        ----------
        binding_site_id : int or list of int, default=None

        Returns
        -------
        lipidcount_BS : list, len(lipidcount_BS)=len(binding_site_id)

        """
        self._check_calculation("Residue", self.collect_residue_contacts)
        self._check_calculation("Binding Site ID", self.compute_binding_nodes, print_data=False)
        selected_bs_id = np.atleast_1d(binding_site_id) if binding_site_id is not None \
            else np.arange(len(self._node_list), dtype=int)
        for bs_id in selected_bs_id:
            nodes = self._node_list[bs_id]
            lipidcount_BS = []
            for traj_idx in np.arange(len(self._trajfile_list), dtype=int):
                for protein_idx in np.arange(self._nprot, dtype=int):
                    list_to_take = traj_idx * self._nprot + protein_idx
                    n_frames = len(self._contact_residues_low[nodes[0]][list_to_take])
                    contact_BS_low = [np.unique(np.concatenate(
                        [self._contact_residues_low[node][list_to_take][frame_idx] for node in nodes]))
                        for frame_idx in np.arange(n_frames)]
                    lipidcount_BS.append(cal_occupancy(contact_BS_low))
            self._lipid_count_BS[bs_id] = lipidcount_BS
        # update dataset
        lipidcount_BS_per_residue = np.zeros(self._nresi_per_protein)
        for bs_id, nodes in enumerate(self._node_list):
            lipidcount_BS_per_residue[nodes] = np.mean(self._lipid_count_BS[bs_id]) \
                if len(self._lipid_count_BS[bs_id]) > 0 else 0
        self.dataset["Binding Site Lipid Count"] = lipidcount_BS_per_residue

        if len(selected_bs_id) == 1:
            return self._lipid_count_BS[bs_id]
        else:
            return [self._lipid_count_BS[bs_id] for bs_id in selected_bs_id]

    def compute_site_koff(self, binding_site_id=None, nbootstrap=10, initial_guess=[1., 1., 1., 1.],
                          save_dir=None, plot_data=True, fig_close=True):
        """Calculate interactions koff and residence time for binding sites.

        Parameters
        ----------
        binding_site_id : int or array_like or None, default=None
        nbootstrap : int, default=10
        initial_guess : array_like, default=None
        save_dir : str, default=None
        plot_data : bool, default=True
        fig_close : bool, default=True

        Returns
        ---------
        koffs : list
        restimes : list

        """
        self._check_calculation("Residue", self.collect_residue_contacts)
        self._check_calculation("Binding Site ID", self.compute_binding_nodes, print_data=False)
        if plot_data:
            BS_dir = check_dir(save_dir, "Binding_Sites_koffs_{}".format(self._lipid)) if save_dir is not None \
                else check_dir(self._save_dir, "Binding_Sites_koffs_{}".format(self._lipid))

        selected_bs_id = np.atleast_1d(binding_site_id) if binding_site_id is not None \
            else np.arange(len(self._node_list), dtype=int)
        binding_sites_missing_durations = [bs_id for bs_id in selected_bs_id
                                           if len(self._duration_BS[bs_id]) == 0]
        if len(binding_sites_missing_durations) > 0:
            self.compute_site_duration(binding_site_id=binding_sites_missing_durations)
        t_total = np.max(self._T_total)
        same_length = np.all(np.array(self._T_total) == t_total)
        if not same_length:
            warnings.warn(
                "Trajectories have different lengths. This will impair the accuracy of koff calculation!")
        timestep = np.min(self._timesteps)
        same_timestep = np.all(np.array(self._timesteps) == timestep)
        if not same_timestep:
            warnings.warn(
                "Trajectories have different timesteps. This will impair the accuracy of koff calculation!")
        # calculate koff for binding sites
        for bs_id in tqdm(selected_bs_id, desc="CALCULATE KOFF FOR BINDING SITES", total=len(selected_bs_id)):
            durations = np.concatenate(self._duration_BS[bs_id])
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                koff_BS, restime_BS, properties_BS = cal_koff(durations, t_total, timestep, nbootstrap, initial_guess)
            self._koff_BS[bs_id] = koff_BS
            self._koff_BS_boot[bs_id] = np.mean(properties_BS["ks_boot_set"], axis=0)[0]
            self._res_time_BS[bs_id] = restime_BS
            self._r_squared_BS[bs_id] = properties_BS["r_squared"]
            self._r_squared_BS_boot[bs_id] = np.mean(properties_BS["r_squared_boot_set"])
            if plot_data:
                # plot BS koff and BS residence time
                text = self._format_koff_text(properties_BS)
                plot_koff(durations, properties_BS["delta_t_list"], properties_BS["survival_rates"],
                          properties_BS["n_fitted"], survival_rates_bootstraps=properties_BS["survival_rates_boot_set"],
                          fig_fn=os.path.join(BS_dir, f"BS_id{bs_id}.pdf"), title=f"Binding Site {bs_id}",
                          timeunit=self._timeunit, t_total=t_total, text=text, fig_close=fig_close)
        # update dataset
        for data, column_name in zip(
                [self._koff_BS, self._koff_BS_boot, self._res_time_BS, self._r_squared, self._r_squared_BS_boot],
                ["Binding Site Koff", "Binding Site Koff Bootstrap avg", "Binding Site Residence Time",
                 "Binding Site R Squared", "Binding Site R Squared Bootstrap avg"]):
            data_per_residue = np.zeros(self._nresi_per_protein)
            for bs_id, nodes in enumerate(self._node_list):
                data_per_residue[nodes] = data[bs_id]
            self.dataset[column_name] = data_per_residue

        if len(selected_bs_id) == 1:
            return self._koff_BS[bs_id], self._res_time_BS[bs_id]
        else:
            return [self._koff_BS[bs_id] for bs_id in selected_bs_id], \
                   [self._res_time_BS[bs_id] for bs_id in selected_bs_id]

    def analyze_bound_poses(self, binding_site_id=None, n_top_poses=3, pose_format="gro", score_weights=None,
                            kde_bw=0.15, pca_component=0.95, plot_rmsd=True, save_dir=None,
                            n_clusters="auto", eps=None, min_samples=None, metric="euclidean", fig_close=False):
        """Analyze bound poses for binding sites.

        This function can find representative bound poses, cluster the bound poses and calculate pose RMSD for
        binding sites.

        Parameters
        ----------
        binding_site_id : int or list_like or None, default=None
        n_top_poses : int, default=5
        pose_format : str, default="pdb"
        score_weights : dict or None, default=None
        kde_bw : scalar, default=0.15
        pca_component : int, float or ‘mle’, default=0.95
            Set the `n_components` value in `sklearn.decomposition.PCA`
        plot_rmsd : bool, default=True
        n_clusters : int or 'auto'
            If `n_clusters == 'auto'`, the number of clusters will be guessed based on density, whereas n_clusters=N
            will use KMeans to cluster the poses.
        save_dir : str or None, default=None

        Returns
        -------
        pose_pool : dict
            Coordinates of all bound poses in stored in a python dictionary {binding_site_id: pose coordinates}
        rmsd_data : pandas.DataFrame
            Bound poses RMSDs are stored by columns with column name of binding site id.

        """
        self._check_calculation("Binding Site ID", self.compute_binding_nodes, print_data=False)

        pose_dir = check_dir(save_dir, "Bound_Poses_{}".format(self._lipid)) if save_dir is not None \
            else check_dir(self._save_dir, "Bound_Poses_{}".format(self._lipid))
        if "Binding Site RMSD" in self.dataset.columns:
            pose_rmsd_per_residue = self.dataset.columns["Binding Site Pose RMSD"]
        else:
            pose_rmsd_per_residue = np.zeros(self._nresi_per_protein)
        if binding_site_id is not None:
            selected_bs_id = np.atleast_1d(binding_site_id)
        else:
            selected_bs_id = np.arange(len(self._node_list), dtype=int)

        # store bound lipid poses
        selected_bs_map = {bs_id: self._node_list[bs_id] for bs_id in selected_bs_id}
        pose_pool = collect_bound_poses(selected_bs_map, self._contact_residues_low, self._trajfile_list,
                                        self._topfile_list, self._lipid, stride=self._stride, nprot=self._nprot)
        # analyize bound poses
        RMSD_set = {}
        for bs_id in tqdm(selected_bs_id, desc="ANALYZE BOUND POSES"):
            # lipid_dist_per_atom shape: [n_lipid_atoms, n_bound_poses, n_BS_residues]
            if len(pose_pool[bs_id]) == 0:
                print(f"Binding Site {bs_id} is bogus! Possibly due to insufficient sampling.")
                continue
            lipid_dist_per_atom, pose_traj = vectorize_poses(pose_pool[bs_id], self._node_list[bs_id],
                                                             self._protein_ref, self._lipid_ref)
            if n_top_poses > 0:
                pose_dir_rank = check_dir(pose_dir, "BSid{}_rank".format(bs_id), print_info=False)
                atom_weights = {atom_idx: 1 for atom_idx in np.arange(self._lipid_ref.n_atoms)}
                if score_weights is not None:
                    translate = {atom_idx: score_weights[self._lipid_ref.atom(atom_idx)]
                                 for atom_idx in np.arange(self._lipid_ref.n_atoms)
                                 if self._lipid_ref.atom(atom_idx) in score_weights.keys()}
                    atom_weights.update(translate)
                scores = calculate_scores(lipid_dist_per_atom, kde_bw=kde_bw, pca_component=pca_component,
                                          score_weights=atom_weights)
                num_of_poses = min(n_top_poses, pose_traj.n_frames)
                pose_indices = np.argsort(scores)[::-1][:num_of_poses]
                write_bound_poses(pose_traj, pose_indices, pose_dir_rank, pose_prefix="BSid{}_top".format(bs_id),
                                  pose_format=pose_format)
            lipid_dist_per_pose = np.array([lipid_dist_per_atom[:, pose_id, :].ravel()
                                            for pose_id in np.arange(lipid_dist_per_atom.shape[1])])
            # cluster poses
            if n_clusters == 'auto':
                pose_dir_clustered = check_dir(pose_dir, "BSid{}_clusters".format(bs_id), print_info=False)
                transformed_data = PCA(n_components=pca_component).fit_transform(lipid_dist_per_pose)
                _, core_sample_indices = cluster_DBSCAN(transformed_data, eps=eps, min_samples=min_samples,
                                                        metric=metric)
                selected_pose_id = [np.random.choice(i_core_sample, 1)[0] for i_core_sample in core_sample_indices]
                write_bound_poses(pose_traj, selected_pose_id, pose_dir_clustered,
                                  pose_prefix="BSid{}_cluster".format(bs_id), pose_format=pose_format)
            elif n_clusters > 0:
                pose_dir_clusters = check_dir(pose_dir, "BSid{}_clusters".format(bs_id), print_info=False)
                transformed_data = PCA(n_components=pca_component).fit_transform(lipid_dist_per_pose)
                cluster_labels = cluster_KMeans(transformed_data, n_clusters=n_clusters)
                cluster_id_set = np.unique(cluster_labels)
                selected_pose_id = [np.random.choice(np.where(cluster_labels == cluster_id)[0], 1)[0]
                                    for cluster_id in cluster_id_set]
                write_bound_poses(pose_traj, selected_pose_id, pose_dir_clusters,
                                  pose_prefix="BSid{}_cluster".format(bs_id), pose_format=pose_format)
            # calculate pose rmsd
            dist_mean = np.mean(lipid_dist_per_pose, axis=0)
            RMSD_set["Binding Site {}".format(bs_id)] = [rmsd(lipid_dist_per_pose[pose_id], dist_mean)
                                                         for pose_id in np.arange(len(lipid_dist_per_pose))]
            pose_rmsd_per_residue[self._node_list[bs_id]] = np.mean(RMSD_set["Binding Site {}".format(bs_id)])
        # update dataset
        self.dataset["Binding Site Pose RMSD"] = pose_rmsd_per_residue
        pose_rmsd_data = pd.DataFrame(
            dict([(bs_label, pd.Series(rmsd_set)) for bs_label, rmsd_set in RMSD_set.items()]))
        # plot RMSD
        if plot_rmsd and n_top_poses > 0:
            plot_binding_site_data(pose_rmsd_data, os.path.join(pose_dir, "Pose_RMSD_violinplot.pdf"),
                                   title="{}".format(self._lipid), ylabel="RMSD (nm)", fig_close=fig_close)
        return pose_pool, pose_rmsd_data

    def compute_surface_area(self, binding_site_id=None, radii=None, plot_data=True, save_dir=None, fig_close=False):
        """Calculate binding site surface areas.

        Parameters
        -----------
        binding_site_id : int or array_like or None, default=None
        radii : dict or None, default=None
        plot_data : bool, default=True
        save_dir : str or None, default=None

        """
        MARTINI_CG_radii = {"BB": 0.26, "SC1": 0.23, "SC2": 0.23, "SC3": 0.23}

        self._check_calculation("Binding Site ID", self.compute_binding_nodes, print_data=False)

        if "Binding Site Surface Area" in self.dataset.columns:
            # keep existing data
            surface_area_per_residue = np.array(self.dataset["Binding Site Surface Area"].tolist())
        else:
            surface_area_per_residue = np.zeros(self._nresi_per_protein)
        if plot_data:
            selected_bs_id = np.atleast_1d(np.array(binding_site_id, dtype=int)) if binding_site_id is not None \
                else np.arange(len(self._node_list), dtype=int)
        if radii is None:
            radii_book = MARTINI_CG_radii
        else:
            radii_book = {**MARTINI_CG_radii, **radii}

        # calculate binding site surface area
        selected_bs_id_map = {bs_id: self._node_list[bs_id] for bs_id in selected_bs_id}
        surface_area_data = calculate_site_surface_area(selected_bs_id_map, radii_book, self._trajfile_list,
                                                        self._topfile_list, self._nprot, self._timeunit, self._stride,
                                                        dt_traj=self._dt_traj)
        # update dataset
        for bs_id in selected_bs_id:
            nodes = self._node_list[bs_id]
            surface_area_per_residue[nodes] = surface_area_data["Binding Site {}".format(bs_id)].mean()
        self.dataset["Binding Site Surface Area"] = surface_area_per_residue
        # plot surface area
        if plot_data:
            if save_dir is not None:
                surface_area_dir = check_dir(save_dir)
            else:
                surface_area_dir = check_dir(self._save_dir)
            plot_surface_area(surface_area_data,
                              os.path.join(surface_area_dir, "Surface_Area_{}_timeseries.pdf".format(self._lipid)),
                              timeunit=self._timeunit, fig_close=fig_close)
            selected_columns = [column for column in surface_area_data.columns if column != "Time"]
            surface_data_noTimeSeries = surface_area_data[selected_columns]
            plot_binding_site_data(surface_data_noTimeSeries,
                                   os.path.join(surface_area_dir, "Surface_Area_{}_violinplot.pdf".format(self._lipid)),
                                   title="{}".format(self._lipid), ylabel=r"Surface Area (nm$^2$)",
                                   fig_close=fig_close)
        return surface_area_data

    #################################
    #    save and plot
    #################################
    def save_data(self, item, save_dir=None):
        """Assisting function for saving data.

        Parameters
        ----------
        item : {"Dataset", "Duration", "Occupancy", "Lipid Count", "CorrCoef", "Duration BS",
                "Occupancy BS"}
        save_dir : str, optional, default=None

        """
        data_dir = check_dir(save_dir, "Dataset_{}".format(self._lipid)) if save_dir is not None \
            else check_dir(self._save_dir, "Dataset_{}".format(self._lipid))

        if item.lower() == "duration":
            obj = self._duration
        elif item.lower() == "occupancy":
            obj = self._occupancy
        elif item.lower() == "lipid count":
            obj = self._lipid_count
        elif item.lower() == "corrcoef":
            obj = self.interaction_corrcoef
        elif item.lower() == "duration bs":
            obj = self._duration_BS
        elif item.lower() == "occupancy bs":
            obj = self._occupancy_BS
        if 'obj' in locals():
            with open(os.path.join(data_dir, "{}.pickle".format("_".join(item.split()))), "wb") as f:
                pickle.dump(obj, f, 2)

        if item.lower() == "dataset":
            self.dataset.to_csv(os.path.join(data_dir, "dataset.csv"), header=True, index=False)

        return

    def save_coordinate(self, item, save_dir=None, fn_coord=None):
        """Save protein coordinates in PDB format with interaction data in the B factor column.

        parameters
        -----------
        item : {"Residence Time", "Duration", "Occupancy", "Lipid Count"}
        save_dir : str or None, default=None

        """
        coord_dir = check_dir(save_dir, "Coordinate_{}".format(self._lipid)) if save_dir is not None \
            else check_dir(self._save_dir, "Coordinate_{}".format(self._lipid))
        if fn_coord is None:
            fn_coord = "Coordinate_{}_{}.pdb".format(self._lipid, "_".join(item.split()))
        data = self.dataset[item].tolist()
        write_PDB(self._protein_ref, data, os.path.join(coord_dir, fn_coord), resi_offset=self._resi_offset)
        return

    def save_pymol_script(self, pdb_file, save_dir=None):
        """save a pymol script that maps interactions onto protein structure in PyMol.

        Parameters
        ----------
        pdb_file : str
        save_dir : str, optional, default=None

        """
        script_dir = check_dir(save_dir) if save_dir is not None else check_dir(self._save_dir)
        data_fname = os.path.join(script_dir, "Dataset_{}.csv".format(self._lipid))
        if not os.path.isfile(data_fname):
            self.dataset.to_csv(data_fname, index=False, header=True)
        write_pymol_script(os.path.join(script_dir, "show_binding_site_info.py"), pdb_file, data_fname,
                           self._lipid, len(self._node_list))
        return

    def plot(self, item, save_dir=None, gap=200, fig_close=False):
        """Assisting function for plotting interaction data.

        Plot interactions per residue or plot interaction correlation matrix.

        Parameters
        ----------
        item : {"Residence Time", "Duration", "Occupancy", "Lipid Count", "CorrCoef"}
        save_dir : str, default=None
        gap : int, default=200

        """
        figure_dir = check_dir(save_dir, "Figure_{}".format(self._lipid)) if save_dir is not None \
            else check_dir(self._save_dir, "Figure_{}".format(self._lipid))

        if item == "Residence Time":
            ylabel = "Res. Time (ns)" if self._timeunit == "ns" else r"Res. Time ($\mu$s)"
        elif item == "Duration":
            ylabel = "Duration (ns)" if self._timeunit == 'ns' else r"Duration ($\mu$s)"
        elif item == "Occupancy":
            ylabel = "Occupancy (%)"
        elif item == "Lipid Count":
            ylabel = "Lipid Count (num.)"
        if "ylabel" in locals():
            data = self.dataset[item]
            title = "{} {}".format(self._lipid, item)
            fig_fn = os.path.join(figure_dir, "{}.pdf".format("_".join(item.split())))
            residue_index = np.array([int(re.findall("^[0-9]+", residue)[0]) for residue in self._residue_list])
            plot_residue_data(residue_index, data, gap=gap, ylabel=ylabel, fn=fig_fn, title=title,
                              fig_close=fig_close)

        if item == "CorrCoef":
            residue_index = np.array([int(re.findall("^[0-9]+", residue)[0]) for residue in self._residue_list])
            plot_corrcoef(self.interaction_corrcoef, residue_index, fn=os.path.join(figure_dir, "CorrCoef.pdf"),
                          title="{} Correlation Coeffient".format(self._lipid), fig_close=fig_close)

        return

    def plot_logo(self, item, save_dir=None, gap=2000, letter_map=None, color_scheme="chemistry", fig_close=False):
        """Plot interactions using logomaker.

        Parameters
        ----------
        item : {"Residence Time", "Duration", "Occupancy", "Lipid Count"}
        save_dir : str or None, optional, default=None
        gap : int, optional, default=2000
        letter_map : dict or None, optional, default=None
        color_scheme : str, optional, default="chemistry"

        """
        figure_dir = check_dir(save_dir, "Figure_{}".format(self._lipid)) if save_dir is not None \
            else check_dir(self._save_dir, "Figure_{}".format(self._lipid))

        single_letter = {'CYS': 'C', 'ASP': 'D', 'SER': 'S', 'GLN': 'Q', 'LYS': 'K',
                         'ILE': 'I', 'PRO': 'P', 'THR': 'T', 'PHE': 'F', 'ASN': 'N',
                         'GLY': 'G', 'HIS': 'H', 'LEU': 'L', 'ARG': 'R', 'TRP': 'W',
                         'ALA': 'A', 'VAL': 'V', 'GLU': 'E', 'TYR': 'Y', 'MET': 'M'}
        if letter_map is not None:
            single_letter.update(letter_map)

        resname_set = [re.findall("[a-zA-Z]+$", residue)[0] for residue in self._residue_list]
        residue_index = np.array([int(re.findall("^[0-9]+", residue)[0]) for residue in self._residue_list])

        if item == "Residence Time":
            ylabel = "Res. Time (ns)" if self._timeunit == "ns" else r"Res. Time ($\mu$s)"
        elif item == "Duration":
            ylabel = "Duration (ns)" if self._timeunit == 'ns' else r"Duration ($\mu$s)"
        elif item == "Occupancy":
            ylabel = "Occupancy (%)"
        elif item == "Lipid Count":
            ylabel = "Lipid Count (num.)"

        if "ylabel" in locals():
            data = self.dataset[item]
            title = "{} {} Logo".format(self._lipid, item)
            fig_fn = os.path.join(figure_dir, "{}_logo.pdf".format("_".join(item.split())))
            plot_residue_data_logos(residue_index, resname_set, data, ylabel=ylabel,
                                    fn=fig_fn, title=title, letter_map=letter_map, color_scheme=color_scheme,
                                    fig_close=fig_close)
        else:
            print("Invalid input for item!")

        return

    def write_site_info(self, sort_residue="Residence Time", save_dir=None, fn=None):
        """Write a report on binding site with lipid interaction information.

        The report contains information of the binding site residence time, durations, occupancy,
        lipid count, surface area and some stats on amino acids composition. A list of lipid interactions with
        comprising residues will follow after the binding site information, in a sorted order provided by
        `sort_residue`.

        Parameters
        ----------
        sort_residue : {"Residence Time", "Duration", "Occupancy", "Lipid Count"}, default="Residence Time"
        save_dir : str, default=None
        fn : str, default=None

        """
        if len(self._node_list) == 0:
            print("No binding site was detected!!")
        else:
            BS_dir = check_dir(save_dir) if save_dir is not None \
                else check_dir(self._save_dir)
            if fn is None:
                fn = "BindingSites_Info_{}.txt".format(self._lipid)
            self._check_calculation("Binding Site ID", self.compute_binding_nodes, print_data=False)
            self._check_calculation("Binding Site Koff", self.compute_site_koff, print_data=False)
            with open(os.path.join(BS_dir, fn), "a") as f:
                f.write(f"## Network modularity {self._network_modularity:5.3f}")
                f.write("\n")
                for bs_id, nodes in enumerate(self._node_list):
                    text = self._format_BS_print_info(bs_id, nodes, sort_residue)
                    f.write(text)
        return

    def show_stats_per_traj(self, write_log=True, print_log=True, fn_log=None):
        """Show stats of interactions per trajectory.

        Parameters
        ----------
        write_log : bool, default=True
        print_log : bool, default=True
        fn_log : str, default=None

        """
        if fn_log is None:
            fn_log = "calculation_log_{}.txt".format(self.lipid)
        text = []
        text.append("###### Lipid: {}\n".format(self._lipid))
        text.append("###### Lipid Atoms: {}\n".format(self._lipid_atoms))
        text.append("###### Cutoffs: {}\n".format(self.cutoffs))
        text.append("###### nprot: {}\n".format(self._nprot))
        text.append("###### Trajectories:\n")
        for traj_fn in self._trajfile_list:
            text.append("  {}\n".format(traj_fn))
        text.append("###### Coordinates:\n")
        for top_fn in self._topfile_list:
            text.append("  {}\n".format(top_fn))
        # interaction information by trajectory
        for traj_idx, trajfile in enumerate(self._trajfile_list):
            # sort values
            durations = np.array(
                [np.concatenate(
                    self._duration[residue_id][traj_idx * self._nprot:(traj_idx + 1) * self._nprot]).mean()
                 for residue_id in self._protein_residue_id])
            duration_arg_idx = np.argsort(durations)[::-1]
            occupancies = np.array(
                [np.mean(self._occupancy[residue_id][traj_idx * self._nprot:(traj_idx + 1) * self._nprot])
                 for residue_id in self._protein_residue_id])
            occupancy_arg_idx = np.argsort(occupancies)[::-1]
            lipid_counts = np.array(
                [np.mean(self._lipid_count[residue_id][traj_idx * self._nprot:(traj_idx + 1) * self._nprot])
                 for residue_id in self._protein_residue_id])
            lipid_count_arg_idx = np.argsort(lipid_counts)[::-1]

            text.append(
                "\n########## {} interactions in \n########## {} \n".format(self._lipid, trajfile))
            text.append(
                "10 residues that showed longest average interaction durations ({}):\n".format(self._timeunit))
            for residue, duration in zip(self._residue_list[duration_arg_idx][:10], durations[duration_arg_idx][:10]):
                text.append("{:^8s} -- {:^8.3f}\n".format(residue, duration))
            text.append("10 residues that showed highest lipid occupancy (100%):\n")
            for residue, occupancy in zip(self._residue_list[occupancy_arg_idx][:10],
                                          occupancies[occupancy_arg_idx][:10]):
                text.append("{:^8s} -- {:^8.2f}\n".format(residue, occupancy))
            text.append("10 residues that have the largest number of surrounding lipids (count):\n")
            for residue, lipid_count in zip(self._residue_list[lipid_count_arg_idx][:10],
                                            lipid_counts[lipid_count_arg_idx][:10]):
                text.append("{:^8s} -- {:^8.2f}\n".format(residue, lipid_count))
            text.append("\n")
            text.append("\n")

        log = "".join(text)

        if write_log:
            with open("{}/{}".format(self._save_dir, fn_log), "a+") as f:
                f.write(log)

        if print_log:
            print(log)

        return

    ############################################
    #     assisting func
    ############################################

    def _format_koff_text(self, properties):
        """Format text for koff plot. """
        tu = "ns" if self._timeunit == "ns" else r"$\mu$s"
        text = "{:18s} = {:.3f} {:2s}$^{{-1}} $\n".format("$k_{{off1}}$", properties["ks"][0], tu)
        text += "{:18s} = {:.3f} {:2s}$^{{-1}} $\n".format("$k_{{off2}}$", properties["ks"][1], tu)
        text += "{:14s} = {:.4f}\n".format("$R^2$", properties["r_squared"])
        ks_boot_avg = np.mean(properties["ks_boot_set"], axis=0)
        cv_avg = 100 * np.std(properties["ks_boot_set"], axis=0) / np.mean(properties["ks_boot_set"], axis=0)
        text += "{:18s} = {:.3f} {:2s}$^{{-1}}$ ({:3.1f}%)\n".format("$k_{{off1, boot}}$", ks_boot_avg[0],
                                                                     tu, cv_avg[0])
        text += "{:18s} = {:.3f} {:2s}$^{{-1}}$ ({:3.1f}%)\n".format("$k_{{off2, boot}}$", ks_boot_avg[1],
                                                                     tu, cv_avg[1])
        text += "{:14s} = {:.4f}\n".format("$R^2$$_{{boot}}$", np.mean(properties["r_squared_boot_set"]))
        text += "{:18s} = {:.3f} {:2s}".format("$Res. Time$", properties["res_time"], tu)
        return text

    def _format_BS_print_info(self, bs_id, nodes, sort_item):
        """Format binding site information."""
        Residue_property_book = {"ARG": "Pos. Charge", "HIS": "Pos. Charge", "LYS": "Pos. Charge",
                                 "ASP": "Neg. Charge", "GLU": "Neg. Charge",
                                 "SER": "Polar", "THR": "Polar", "ASN": "Polar", "GLN": "Polar",
                                 "CYS": "Special", "SEC": "Special", "GLY": "Special", "PRO": "Special",
                                 "ALA": "Hydrophobic", "VAL": "Hydrophobic", "ILE": "Hydrophobic", "LEU": "Hydrophobic",
                                 "MET": "Hydrophobic", "PHE": "Hydrophobic", "TYR": "Hydrophobic", "TRP": "Hydrophobic"}
        text = "# Binding site {}\n".format(bs_id)
        if "Binding Site Residence Time" in self.dataset.columns:
            text += "{:30s} {:10.3f} {:5s}\n".format(" Binding Site Residence Time:",
                                                     self._res_time_BS[bs_id], self._timeunit)
        if "Binding Site R Squared" in self.dataset.columns:
            text += "{:30s} {:10.3f}  R squared: {:7.4f}\n".format(" Binding Site Koff:", self._koff_BS[bs_id],
                                                                   self._r_squared_BS[bs_id])
        if "Binding Site Duration" in self.dataset.columns:
            text += "{:30s} {:10.3f} {:5s}\n".format(" Binding Site Duration:",
                                                     np.concatenate(self._duration_BS[bs_id]).mean(), self._timeunit)
        if "Binding Site Occupancy" in self.dataset.columns:
            text += "{:30s} {:10.3f} %\n".format(" Binding Site Occupancy:",
                                                 np.mean(self._occupancy_BS[bs_id]))
        if "Binding Site Lipid Count" in self.dataset.columns:
            text += "{:30s} {:10.3f}\n".format(" Binding Site Lipid Count:",
                                               np.mean(self._lipid_count_BS[bs_id]))
        if "Binding Site Surface Area" in self.dataset.columns:
            text += "{:30s} {:10.3f} nm^2\n".format(" Binding Site Surface Area:",
                            np.mean(self.dataset[self.dataset["Binding Site ID"]==bs_id]["Binding Site Surface Area"]))
        res_stats = {"Pos. Charge": 0, "Neg. Charge": 0, "Polar": 0, "Special": 0, "Hydrophobic": 0}
        # stats on the chemical properties of binding site residues
        for residue in self._residue_list[nodes]:
            res_stats[Residue_property_book[re.findall("[a-zA-Z]+$", residue)[0]]] += 1
        BS_num_resi = len(nodes)
        text += "{:20s} {:10s}\n".format(" Pos. Charge:", "/".join([str(res_stats["Pos. Charge"]), str(BS_num_resi)]))
        text += "{:20s} {:10s}\n".format(" Neg. Charge:", "/".join([str(res_stats["Neg. Charge"]), str(BS_num_resi)]))
        text += "{:20s} {:10s}\n".format(" Polar:", "/".join([str(res_stats["Polar"]), str(BS_num_resi)]))
        text += "{:20s} {:10s}\n".format(" Hydrophobic:", "/".join([str(res_stats["Hydrophobic"]), str(BS_num_resi)]))
        text += "{:20s} {:10s}\n".format(" Special:", "/".join([str(res_stats["Special"]), str(BS_num_resi)]))
        text += "{:^9s}{:^7s}{:^16s}{:^16s}{:^15s}{:^12s}{:^7s}{:^10s}\n".format("Residue", "Res ID",
                                                                                 "Res. Time ({})".format(
                                                                                     self._timeunit),
                                                                                 "Duration ({})".format(self._timeunit),
                                                                                 "Occupancy (%)", "Lipid Count",
                                                                                 "Koff", "R Squared")
        info_dict_set = self.dataset.iloc[nodes].sort_values(by=sort_item, ascending=False).to_dict("records")
        for info_dict in info_dict_set:
            info_dict_converted = defaultdict(float, info_dict)
            text += "{Residue:^9s}{Residue ID:^7d}{Residence Time:^16.3f}{Duration:^16.3f}" \
                    "{Occupancy:^15.3f}{Lipid Count:^12.3f}{Koff:^7.3f}{R Squared:^10.3f}\n".format(**info_dict_converted)
        text += "\n"
        text += "\n"
        return text

    def _check_calculation(self, item, calculation, *args, **kwargs):
        """Check BS calculation. """
        if item not in self.dataset.columns:
            print("#" * 60)
            print(f"{item} has not been calculated. Start the calculation.")
            calculation(*args, **kwargs)
            print("Finished the calculation of {}.".format(item))
            print("#" * 60)



##############################################################################
# PyLipID: A python module for analysing protein-lipid interactions
#
# Author: Wanling Song
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
##############################################################################

"""This module contains the assisting functions for dealing with directories."""

import os

__all__ = ["check_dir"]


def check_dir(directory=None, suffix=None, print_info=True):
    """Creating directory

    This function will combine the suffix with the given directory (or the current
    working directory if none is given) to generate a directory name, and create a
    directory with this name if it does not exit.

    """
    if directory is None:
        directory = os.getcwd()
    else:
        directory = os.path.abspath(directory)
    if suffix is not None:
        directory = os.path.join(directory, suffix)
    if not os.path.isdir(directory):
        os.makedirs(directory)
        if print_info:
            print("Creating new director: {}".format(directory))
    return directory




##############################################################################
# PyLipID: A python module for analysing protein-lipid interactions
#
# Author: Wanling Song
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
##############################################################################

"""This module contains functions for calculating interaction residence time and koff.
"""
import numpy as np
from scipy.optimize import curve_fit


__all__ = ["cal_koff", "cal_survival_func"]


def cal_koff(durations, t_total, timestep, nbootstrap=10, initial_guess=[1., 1., 1., 1.], cap=True):
    """Calculate residence time based on collected interaction durations via a normalised survival function.

    Parameters
    ----------
    durations : array_like
            Collected interaction durations
    t_total : scalar
            The duration or length, or the longest if using multiple simulations of different durations/lengths, of the
            simulation trajectories. Should be in the same time unit as durations.
    timestep : scalar
            :math:`\Delta t` of the survival function :math:`\sigma`. Often take the time step of the simulation
            trajectories or multiples of the trajectory time step. Should be in the same time unit as durations.
    nbootstrap : int, optional, default=10
            Number of bootstrapping for calculation. The default is 10.
    initial_quess : list, optional, default=(1., 1., 1., 1.)
            The initial guess for bi-exponential fitting to the survival function that is used by
            scipy.optimize.curve_fit.
    cap : bool, optional, default=True
            Cap the returned residence time to t_total.

    Returns
    ----------
    koff : scalar
            The calculated koff. In the unit of "tu"^-1 in which "tu" is the time unit used by the provided durations.
    res_time : scalar
            The calculated residence time. In the same time unit as the provided durations.
    properties : dict
            A dictionary of all the computed values, including the original and bootstrapped koffs, residence times, ks 
            of the bi-expo curve :math:`y=A*e^{(-k_1*x)}+B*e^{(-k_2*x)}` and :math:`R^2`.

    Notes
    ----------
    See `cal_survival_func` for the definition of the survival function :math:`\sigma`.
    A bi-exponential :math:`y=Ae^{-k_1\Delta t}+Be^{-k_2\Delta t}` is used to fit the survival rate curve. The two `ks`
    are stored in the returned dictionary of properties. The smaller k is regarded as the interaction koff and
    the residence time is calculated as :math:`{\frac{1}{koff}}`.

    See also
    -----------
    pylipid.plot.plot_koff
            Plotting function for interaction durations and the calculated survival function.

    """
    # calculate original residence time
    delta_t_list = np.arange(0, t_total, timestep)
    survival_func = cal_survival_func(durations, np.max(t_total), delta_t_list)
    survival_rates = np.array([survival_func[delta_t] for delta_t in delta_t_list])
    res_time, _, r_squared, params = _curve_fitting(survival_func, delta_t_list, initial_guess)
    if cap and res_time > t_total:
        res_time = t_total
    n_fitted = _bi_expo(np.array(delta_t_list), *params)
    r_squared = 1 - np.sum((np.nan_to_num(n_fitted) - np.nan_to_num(survival_rates)) ** 2) / np.sum(
        (survival_rates - np.mean(survival_rates)) ** 2)
    ks = [abs(k) for k in params[:2]]
    ks.sort() # the smaller k is considered as koff

    # calculate bootstrapped residence time
    if nbootstrap > 0:
        duration_boot_set = [np.random.choice(durations, size=len(durations)) for dummy in range(nbootstrap)]
        ks_boot_set = []
        r_squared_boot_set = []
        survival_rates_boot_set = []
        n_fitted_boot_set = []
        for duration_boot in duration_boot_set:
            survival_func_boot = cal_survival_func(duration_boot, np.max(t_total), delta_t_list)
            survival_rates_boot = np.array([survival_func_boot[delta_t] for delta_t in delta_t_list])
            _, _, r_squared_boot, params_boot = _curve_fitting(survival_func_boot, delta_t_list,
                                                                                                initial_guess)
            n_fitted_boot = _bi_expo(np.array(delta_t_list), *params_boot)
            r_squared_boot = 1 - np.sum((np.nan_to_num(n_fitted_boot) - np.nan_to_num(survival_rates_boot)) ** 2) / np.sum(
                (survival_rates_boot - np.mean(survival_rates_boot)) ** 2)
            ks_boot = [abs(k) for k in params_boot[:2]]
            ks_boot.sort()
            ks_boot_set.append(ks_boot)
            r_squared_boot_set.append(r_squared_boot)
            survival_rates_boot_set.append(survival_rates_boot)
            n_fitted_boot_set.append(n_fitted_boot)
    else:
        ks_boot_set = [0]
        r_squared_boot_set = [0]
        survival_rates_boot_set = [0]
        n_fitted_boot_set = [0]

    properties = {"ks": ks, "res_time": res_time, "delta_t_list": delta_t_list,
                  "survival_rates": survival_rates, "survival_rates_boot_set": survival_rates_boot_set,
                  "n_fitted": n_fitted, "n_fitted_boot_set": n_fitted_boot_set,
                  "ks_boot_set": ks_boot_set,
                  "r_squared": r_squared, "r_squared_boot_set": r_squared_boot_set}

    return ks[0], res_time, properties


def cal_survival_func(durations, t_total, delta_t_list):
    """Compute the normalised survival function :math:`\sigma` based on the given durations

    Parameters
    -----------
    durations : array_like
            Collected interaction durations
    t_total : scalar
            The duration or length, or the longest if using multiple simulations of different durations/lengths, of the
            simulation trajectories. Should be in the same time unit as durations.
    delta_t_list : array_like
            The list of :math:`\Delta t` for the survival function :math:`\sigma` to check the interaction survival rate.

    Returns
    -----------
    survival_func : dict
            The survival function :math:`\sigma` stored in a dictionary {delta_t: survival rate}.

    Notes
    -----------
    The normalised survival function

    .. math::
            \sigma\left(t\right)=\frac{1}{N_j}\frac{1}{T-t}\sum_{j=1}^{N_j}\sum_{\nu=0}^{T}{{\tilde{n}}_j\left(\nu,v+t\right)}

    where `T` is the duration of simulation trajectories (`t_total`), :math:`N_j` the number of interactions collected
    from simulations (i.e. len(durations)), and :math:`{\tilde{n}}_j(\nu,v+\Delta t)` is a function that takes the
    value 1 if an interaction appeared for a continuous duration of :math:`\Delta t` after forming the contact at time v,
    and takes the value 0 if otherwise.

    """
    num_of_contacts = len(durations)
    survival_func = {}
    for delta_t in delta_t_list:
        if delta_t == 0:
            survival_func[delta_t] = 1
            survival_func0 = float(sum([res_time - delta_t for res_time in durations if res_time >= delta_t])) / \
                     ((t_total - delta_t) * num_of_contacts)
        else:
            try:
                survival_func[delta_t] = float(sum([res_time - delta_t for res_time in durations if res_time >= delta_t])) / \
                                 ((t_total - delta_t) * num_of_contacts * survival_func0)
            except ZeroDivisionError:
                survival_func[delta_t] = 0
    return survival_func


def _curve_fitting(survival_func, delta_t_list, initial_guess):
    """Fit the exponential curve :math:`y=Ae^{-k_1\Delta t}+Be^{-k_2\Delta t}`"""
    survival_rates = np.nan_to_num([survival_func[delta_t] for delta_t in delta_t_list]) # y
    try:
        popt, pcov = curve_fit(_bi_expo, np.array(delta_t_list), np.array(survival_rates), p0=initial_guess, maxfev=100000)
        n_fitted = _bi_expo(np.array(delta_t_list, dtype=np.float128), *popt)
        r_squared = 1 - np.sum((np.nan_to_num(n_fitted) -
                                np.nan_to_num(survival_rates))**2)/np.sum((survival_rates - np.mean(survival_rates))**2)
        ks = [abs(k) for k in popt[:2]]
        koff = np.min(ks)
        res_time = 1/koff
    except RuntimeError:
        koff = 0
        res_time = 0
        r_squared = 0
        popt = [0, 0, 0, 0]
    return res_time, koff, r_squared, popt


def _bi_expo(x, k1, k2, A, B):
    """The exponential curve :math:`y=Ae^{-k_1\Delta t}+Be^{-k_2\Delta t}`"""
    return A*np.exp(-k1*x) + B*np.exp(-k2*x)



##############################################################################
# PyLipID: A python module for analysing protein-lipid interactions
#
# Author: Wanling Song
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
##############################################################################

"""This module contains functions for calculating correlation coefficient. """

from scipy import sparse
import numpy as np

__all__ = ["sparse_corrcoef"]


def sparse_corrcoef(A, B=None):
    """Calculate correlation coeffient matrix using sparse matrix"""
    if B is not None:
        A = sparse.vstack((A, B), format='csr')
    A = A.astype(np.float64)
    n = A.shape[1]
    # Compute the covariance matrix
    rowsum = A.sum(1)
    centering = rowsum.dot(rowsum.T.conjugate()) / n
    C = (A.dot(A.T.conjugate()) - centering) / (n - 1)
    # The correlation coefficients are given by
    # C_{i,j} / sqrt(C_{i} * C_{j})
    d = np.diag(C)
    corrcoefs = C / np.sqrt(np.outer(d, d))
    return corrcoefs


##############################################################################
# PyLipID: A python module for analysing protein-lipid interactions
#
# Author: Wanling Song
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
##############################################################################

"""This module contains functions for plotting koff.
"""

import os
import numpy as np
import matplotlib.pyplot as plt

__all__ = ["plot_koff"]


def plot_koff(durations, delta_t_list, survival_rates, n_fitted,
              survival_rates_bootstraps=None, fig_fn=None, title=None,
              timeunit=None, text=None, t_total=None, fig_close=True):
    """Plot the koff figure.

    The koff figure contains two axes. The left axis plot the sorted
    interaction durations, and the right one plot normalised survival rates.

    Parameters
    ----------
    durations : array_like
            Collected interaction durations.
    delta_t_list : array_like
            List of :math:`\Delta t` at which the survival rates are calculated.
    survival_rates : array_like
            Survival rates calculated at delta_t_list.
    n_fitted: array_like
            The values of fitted bi-exponential at delta_t_list.
    survival_rates_bootstraps : list of array, optional, default=None
            List of bootstrapped survival rates.
    fig_fn : str, optional, default=None
            Name of the koff figure. by default the figure will be saved as "koff.png"
            in the current working directory.
    title : str, optional, default=None
            Figure title. Default is None.
    t_total : scalar, optional, default=None
            Duration of simulation trajectories. The xlim of both axes will set to t_total if
            a value is given, otherwise xlim will be determined by matplotlib.
    timeunit : {"ns", "us", None}, optional, default=None
            Time unit of the given durations. Default is None.
    text : str, optional, default=None
            Text printed next to the koff figure. The default is None.


    """
    # plot settings
    if timeunit is None:
        xlabel = "Duration (timeunit)"
    elif timeunit == "ns":
        xlabel = "Duration (ns)"
    elif timeunit == "us":
        xlabel = r"Duration ($\mu s$)"

    if text is None:
        fig = plt.figure(1, figsize=(5.5, 3.5))
        left, width = 0.13, 0.33
        bottom, height = 0.17, 0.68
        left_h = left + width + 0.05
        rect_scatter = [left, bottom, width, height]
        rect_histy = [left_h, bottom, width, height]
        axScatter = fig.add_axes(rect_scatter)
        axHisty = fig.add_axes(rect_histy)
    else:
        fig = plt.figure(1, figsize=(8.2, 3.5))
        left, width = 0.0975, 0.23
        bottom, height = 0.17, 0.68
        left_h = left + width + 0.0375
        rect_scatter = [left, bottom, width, height]
        rect_histy = [left_h, bottom, width, height]
        axScatter = fig.add_axes(rect_scatter)
        axHisty = fig.add_axes(rect_histy)

    # plot original data
    x = np.sort(durations)
    y = np.arange(len(x)) + 1
    axScatter.scatter(x[::-1], y, label="Contacts", s=10, c="#176BA0")
    axScatter.set_xlim(0, x[-1] * 1.1)
    axScatter.legend(loc="upper right", prop={"size": 10, "weight": "bold"}, frameon=False, handletextpad=0.1)
    axScatter.set_ylabel("Sorted Index", fontsize=10, weight="bold")
    axScatter.set_xlabel(xlabel, fontsize=10, weight="bold")
    # plot survival function
    axHisty.scatter(delta_t_list, survival_rates, zorder=8, s=10, label="Survival func.", c="#7a5195")
    axHisty.yaxis.set_label_position("right")
    axHisty.yaxis.tick_right()
    axHisty.set_xlabel(r"$\Delta$t", fontsize=10, weight="bold")
    axHisty.set_ylabel("Probability", fontsize=10, weight="bold")
    axHisty.set_yticks([0, 0.25, 0.5, 0.75, 1.0])
    axHisty.set_ylim(-0.1, 1.1)
    # plot the fitted curve
    axHisty.plot(delta_t_list, n_fitted, 'r--', linewidth=3, zorder=10, label="Fitted biexpo.")
    # plot bootstrapped survival functions
    if survival_rates_bootstraps is not None:
        for boot_idx, survival_rates_boot in enumerate(np.atleast_2d(survival_rates_bootstraps)):
            if boot_idx == 0:
                axHisty.plot(delta_t_list, survival_rates_boot, color="gray", alpha=0.5,
                             label="Bootstrapping", linewidth=3)
            else:
                axHisty.plot(delta_t_list, survival_rates_boot, color="gray", alpha=0.5, linewidth=3)

    axHisty.legend(loc="upper right", prop={"size": 8, "weight": "bold"}, frameon=False)
    # set xlim
    if t_total is not None:
        axScatter.set_xlim(0, t_total)
        axHisty.set_xlim(0, t_total)

    if title is not None:
        fig.text(0.13, 0.89, title, fontdict={"size":12, "weight": "bold"})

    # set ticklabel fonts
    for ax in [axHisty, axScatter]:
        for label in ax.xaxis.get_ticklabels() + ax.yaxis.get_ticklabels():
            plt.setp(label, fontsize=10, weight="bold")

    # print text on the right
    axHisty.text(1.4, 1.0, text, verticalalignment='top', horizontalalignment='left', transform=axHisty.transAxes,
                 fontdict={"size": 8, "weight": "normal"}, linespacing=2)

    if fig_fn is None:
        fig_fn = os.path.join(os.getcwd(), "koff.pdf")
    fig.savefig(fig_fn, dpi=300)
    if fig_close:
        plt.close()

    return




# setup.py
from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.rst').read_text(encoding='utf-8')

# read version info
import re
VERSIONFILE="pylipid/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

# setup
setup(
    name='pylipid',
    version=verstr,
    description='PyLipID - A Python Library For Lipid Interaction Analysis',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/wlsong/PyLipID',
    author='Wanling Song',
    author_email='wanling.song@hotmail.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='simulation tools, network community, binding site',
    python_requires='>=3.6, <4',
    packages=find_packages(),
    install_requires=[
        "mdtraj<=1.9.4",
        "numpy",
        "pandas",
        "matplotlib>=3.3.3",
        "networkx",
        "scipy",
        "python-louvain",
        "logomaker",
        "statsmodels",
        "scikit-learn",
        "tqdm",
        "kneebow"
    ]
)



##############################################################################
# PyLipID: A python module for analysing protein-lipid interactions
#
# Author: Wanling Song
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
##############################################################################

"""This module contains functions for calculating binding sites"""

import os
from collections import defaultdict
from itertools import product
import community
import networkx as nx
import numpy as np
import pandas as pd
from tqdm import trange
import mdtraj as md
from statsmodels.nonparametric.kernel_density import KDEMultivariate as kde
from sklearn.decomposition import PCA


__all__ = ["get_node_list", "collect_bound_poses", "vectorize_poses", "calculate_scores", "write_bound_poses",
           "calculate_site_surface_area"]


def get_node_list(corrcoef, threshold=4):
    """Calculate binding sites from correlation coeffient matrix.

    Binding sites are calculated as the community structures of the network built
    on the correlation coefficient matrix.

    Parameters
    -----------
    corrcoef : ndarray(n, n)
    threshold : int, default=4

    Returns
    --------
    node_list : list of lists
    modularity : float or None

    """
    # TODO: check negative values in corrcoef_matrix. Come up with better solutions.
    corrcoef[corrcoef < 0.0] = 0.0 # network edge can't take negative values. Residues with
                                   # negative correlationsare are forced to separate to different binding sites.
    graph = nx.Graph(corrcoef)
    partition = community.best_partition(graph, weight='weight')
    values = [partition.get(node) for node in graph.nodes()]
    node_list = []
    for value in range(max(values)):
        nodes = [k for k, v in partition.items() if v == value]
        if len(nodes) >= threshold:
            node_list.append(nodes)
    if len(node_list) > 0:
        modularity = community.modularity(partition, graph)
    else:
        modularity = None
    return node_list, modularity


def collect_bound_poses(binding_site_map, contact_residue_index, trajfile_list, topfile_list,
                        lipid, stride=1, nprot=1):
    """Collected the bound poses based on contact_residue_index.

    Parameters
    ----------
    binding_site_map : dict
    contact_residue_index : dict
    trajfile_list : list of str
    topfile_list : list of str
    lipid : str
    stride : int,default=1
    nprot : int, default=1

    Returns
    ---------
    pose_pool : dict
        Coordinates of all bound poses in stored in a python dictionary {binding_site_id: pose coordinates}

    See also
    --------
    pylipid.func.cal_contact_residues
        The function that calculates the contact residues from distance matrix

    """
    trajfile_list = np.atleast_1d(trajfile_list)
    topfile_list = np.atleast_1d(topfile_list)
    stride = int(stride)
    nprot = int(nprot)
    pose_pool = defaultdict(list)
    for traj_idx in np.arange(len(trajfile_list)):
        traj = md.load(trajfile_list[traj_idx], top=topfile_list[traj_idx], stride=stride)
        protein_indices_all = traj.top.select("protein")
        natoms_per_protein = int(len(protein_indices_all) / nprot)
        # obtain lipid residue index
        lipid_residue_index_set = set()
        for atom_idx in traj.top.select("resn {}".format(lipid)):
            lipid_residue_index_set.add(traj.top.atom(atom_idx).residue.index)
        lipid_residue_index_set = list(lipid_residue_index_set)
        lipid_residue_index_set.sort()
        # store collect poses
        for protein_idx in np.arange(nprot):
            protein_indices = protein_indices_all[
                              protein_idx * natoms_per_protein:(protein_idx + 1) * natoms_per_protein]
            for bs_id, nodes in binding_site_map.items():
                list_to_take = traj_idx * nprot + protein_idx
                contact_BS = [np.unique(np.concatenate(
                    [contact_residue_index[node][list_to_take][frame_idx] for node in nodes]))
                    for frame_idx in range(traj.n_frames)]
                for frame_id in range(len(contact_BS)):
                    if len(contact_BS[frame_id]) > 0:
                        for lipid_id in contact_BS[frame_id]:
                            lipid_residue_index = lipid_residue_index_set[int(lipid_id)]
                            lipid_atom_indices = np.sort(
                                [atom.index for atom in traj.top.residue(lipid_residue_index).atoms])
                            pose_pool[bs_id].append(
                                [np.copy(traj.xyz[frame_id, np.hstack([protein_indices, lipid_atom_indices])]),
                                 np.copy(traj.unitcell_angles[frame_id]), np.copy(traj.unitcell_lengths[frame_id])])
        return pose_pool


def vectorize_poses(bound_poses, binding_nodes, protein_ref, lipid_ref):
    """Calculate distance matrix for bound poses.

    Parameters
    ----------
    bound_poses :
    binding_nodes :
    protein_ref :
    lipid_ref :

    Returns
    -------
    dist_matrix :
    pose_traj : mdtraj.Trajectory

    """
    # prepare topology
    protein_atom_indices = [[atom.index for atom in residue.atoms] for residue in protein_ref.top.residues]
    lipid_atoms = [protein_ref.n_atoms + atom_idx for atom_idx in np.arange(lipid_ref.n_atoms)]
    joined_top = protein_ref.top.join(lipid_ref.top)

    # start pose generation
    pose_traj = md.Trajectory([frame[0] for frame in bound_poses], joined_top,
                             time=np.arange(len(bound_poses)),
                             unitcell_angles=[frame[1] for frame in bound_poses],
                             unitcell_lengths=[frame[2] for frame in bound_poses])
    # calculate distance to binding site residues for each lipid atom
    dist_per_atom = np.array(
        [np.array([md.compute_distances(pose_traj,
                                        list(product([lipid_atoms[idx]], protein_atom_indices[node])),
                                        periodic=True).min(axis=1) for node in binding_nodes]).T
         for idx in np.arange(lipid_ref.n_atoms)])  # shape: [n_lipid_atoms, n_poses, n_BS_residues]
    return dist_per_atom, pose_traj


def calculate_scores(data, kde_bw=0.15, pca_component=0.95, score_weights=None):
    """Calculate scores based on probability density.

    Parameters
    ----------
    data : ndarray, shape=(n_atoms, n_samples, n_dims)
    kde_bw : scalar
    pca_component : scalar
    score_weights : None or dict
        A dictionary that contains the weight for n_dims, {n_dim: weight}

    Returns
    ---------
    scores : ndarray, shape=(n_samples,)

    See also
    --------
    pylipid.func.collect_bound_poses
    pylipid.func.vectorize_poses

    """
    weights = {atom_idx: 1 for atom_idx in np.arange(np.shape(data)[0])}
    if score_weights is not None:
        weights.update(score_weights)

    kde_funcs = {}
    try:
        for atom_idx in np.arange(np.shape(data)[0]):
            transformed_data = PCA(n_components=pca_component).fit_transform(data[atom_idx])
            var_type = ""
            bw = []
            for dummy in range(len(transformed_data[0])):
                var_type += "c"
                bw.append(kde_bw)
            kde_funcs[atom_idx] = kde(data=transformed_data, var_type=var_type, bw=bw)
        # evaluate binding poses
        scores = np.sum([weights[atom_idx] * kde_funcs[atom_idx].pdf()
                         for atom_idx in np.arange(np.shape(data)[0])], axis=0)
        return scores

    except ValueError:
        print("Pose generation error -- possibly due to insufficient number of binding event.")


def write_bound_poses(pose_traj, pose_indices, save_dir, pose_prefix="BoundPose", pose_format="pdb"):
    """Write bound poses specified by pose_indices.

    Parameters
    ----------
    pose_traj : mdtraj.Trajectory
    pose_indices : array_like
    save_dir : str
    pose_prefix : str, optional, default=BoundPose"
    pose_format : str, optional, default="gro"

    """
    for idx, pose_id in enumerate(pose_indices):
        pose_traj[pose_id].save(os.path.join(save_dir, "{}{}.{}".format(pose_prefix, idx, pose_format)))
    return


def calculate_site_surface_area(binding_site_map, radii_book, trajfile_list, topfile_list,
                                nprot, timeunit, stride, dt_traj=None):
    """Calculate the surface area (in unit of nm^2) as a function of time.

    Parameters
    ----------
    binding_site_map : dict
    trajfile : str or list of str
    topfile : str or list of str
    nprot : int
    timeunit : str, {ns, us}
    stride : int
    dt_traj : scalar or None, optional, default=None

    Returns
    -------
    surface_area : pandas.DataFrame

    """
    surface_data = []
    data_keys = []
    for traj_idx in trange(len(trajfile_list), desc="CALCULATE SURFACE AREA PER TRAJ", total=len(trajfile_list)):
        trajfile, topfile = trajfile_list[traj_idx], topfile_list[traj_idx]
        traj = md.load(trajfile, top=topfile, stride=stride)
        if dt_traj is None:
            traj_times = traj.time / 1000000.0 if timeunit == "us" else traj.time / 1000.0
        else:
            traj_times = float(dt_traj * stride) * np.arange(traj.n_frames)
        protein_indices_all = traj.top.select("protein")
        natoms_per_protein = int(len(protein_indices_all) / nprot)
        for protein_idx in np.arange(nprot):
            protein_indices = protein_indices_all[
                              protein_idx * natoms_per_protein:(protein_idx + 1) * natoms_per_protein]
            new_traj = traj.atom_slice(protein_indices, inplace=False)
            area_all = md.shrake_rupley(new_traj, mode='residue', change_radii=radii_book)
            selected_data = [area_all[:, nodes].sum(axis=1) for bs_id, nodes in binding_site_map.items()]
            selected_data.append(traj_times)
            column_names = ["Binding Site {}".format(bs_id) for bs_id, nodes in binding_site_map.items()]
            column_names.append("Time")
            surface_data.append(pd.DataFrame(np.array(selected_data).T, columns=column_names))
            data_keys.append((traj_idx, protein_idx))
    surface_area = pd.concat(surface_data, keys=data_keys)
    return surface_area



##############################################################################
# PyLipID: A python module for analysing protein-lipid interactions
#
# Author: Wanling Song
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
##############################################################################

"""This module contains functions for writing coordinates in PDB format"""

import numpy as np

__all__ = ["write_PDB"]

def write_PDB(prot_obj, bfactor, pdb_fn, resi_offset=0):
    """Write interaction data in bfactor columns.

    Parameters
    ----------
    prot_obj : mdtraj.TrajectoryObject
    bfactor : array_like
    pdb_fn : str
    resi_offset : int, optional, default=0

    """
    coords = prot_obj.xyz[0]
    table = prot_obj.top.to_dataframe()[0]
    atom_idx_set = table.serial
    resid_set = table.resSeq + resi_offset
    atom_name_set = table.name
    resn_set = table.resName
    chainID = [chr(65 + int(idx)) for idx in table.chainID]
    atom_residue_map = {atom_idx: prot_obj.top.atom(atom_idx).residue.index
                        for atom_idx in np.arange(prot_obj.n_atoms)}
    ######## write out coords ###########
    with open(pdb_fn, "w") as f:
        for idx in np.arange(prot_obj.n_atoms):
            coords_dictionary = {"HEADER": "ATOM",
                                 "ATOM_ID": atom_idx_set[idx],
                                 "ATOM_NAME": atom_name_set[idx],
                                 "SPARE": "",
                                 "RESN": resn_set[idx],
                                 "CHAIN_ID": chainID[idx],
                                 "RESI": resid_set[idx],
                                 "COORDX": coords[idx, 0] * 10,
                                 "COORDY": coords[idx, 1] * 10,
                                 "COORDZ": coords[idx, 2] * 10,
                                 "OCCUP": 1.0,
                                 "BFACTOR": bfactor[atom_residue_map[idx]]}
            row = "{HEADER:6s}{ATOM_ID:5d} ".format(**coords_dictionary) + \
                  "{ATOM_NAME:^4s}{SPARE:1s}{RESN:3s} ".format(**coords_dictionary) + \
                  "{CHAIN_ID:1s}{RESI:4d}{SPARE:1s}   ".format(**coords_dictionary) + \
                  "{COORDX:8.3f}{COORDY:8.3f}{COORDZ:8.3f}{OCCUP:6.2f}{BFACTOR:6.2f}\n".format(**coords_dictionary)
            f.write(row)
        f.write("TER")

    return